//
// GStack.C - implementation of a fast and memory efficient
//            generic stack class which stores pointers to objects.
//            The user must subclass to make a version specific to
//            the type of object they wish to store.  Details can be
//            found in the test program at the bottom of this file.
//            This should use templates, to bad so many people still
//            don't have them. 
//
// $Id: GStack.C,v 1.8 1993/06/20 01:51:53 lijewski Exp $
//
// Copyright 1991, 1992, 1993 by Mike Lijewski.
//

#include <stdlib.h>
#include <string.h>

#include "GStack.h"


GStack::GStack (void) 
{
    blocks = 0;
    index  = -1;
    offset = 0;
    nentries = size = 0;
}


GStack::~GStack (void)
{
    for (int i = 0; i <= index; i++)
        delete [] blocks[i];
    delete [] blocks;
}


GStack::GStack (const GStack& rhs)
{
    index    = rhs.index;
    offset   = rhs.offset;
    nentries = rhs.nentries;
    size     = rhs.size;

    blocks = ::new void**[size];

    for (int i = 0; i <= index; i++)
    {
        blocks[i] = ::new void*[BlockSize];
        memcpy(blocks[i], rhs.blocks[i], BlockSize * sizeof(void*));
    }
}


void GStack::operator= (const GStack& rhs)
{
    if (this == &rhs) return;

    offset   = rhs.offset;
    nentries = rhs.nentries;

    int minindex = index < rhs.index ? index : rhs.index;

    if (size < rhs.size)
    {
        void*** nblocks;
        //
        // We must allocate more space for void** pointers.
        //
        nblocks = ::new void**[rhs.size];
        //
        // Reuse as many of the previous blocks as possible.
        //
        memcpy(nblocks, blocks, (minindex + 1) * sizeof(void**));

        if (minindex == rhs.index)
        {
            //
            // Delete extraneous blocks.
            //
            for (int i = minindex + 1; i <= index; i++)
                delete [] blocks[i];
        }
        else
        {
            //
            // Must allocate new blocks.
            //
            for (int i = minindex + 1; i <= rhs.index; i++)
                nblocks[i] = ::new void*[BlockSize];
        }

        size = rhs.size;
        delete [] blocks;
        blocks = nblocks;
    }
    else
    {
        if (minindex == rhs.index)
        {
            //
            // Delete extraneous blocks.
            //
            for (int i = minindex + 1; i <= index; i++)
                delete [] blocks[i];
        }
        else
        {
            //
            // Must allocate new blocks.
            //
            for (int i = minindex + 1; i <= rhs.index; i++)
                blocks[i] = ::new void*[BlockSize];
        }
    }

    index = rhs.index;

    //
    // Now copy over the void* pointers.
    //
    for (int i = 0; i <= index; i++)
        memcpy(blocks[i], rhs.blocks[i], BlockSize * sizeof(void*));
}


void GStack::push (void* object)
{
    nentries++;

    if (offset == 0)
    {
        //
        // Must add a new block.
        //
        if (++index == 0 || index == size)
        {
            //
            // Must allocate room for more block pointers.
            //
            void*** nblocks = ::new void**[size += ArraySize];
            memcpy(nblocks, blocks, index * sizeof(void**));
            delete [] blocks;
            blocks = nblocks;
        }
        blocks[index] = ::new void*[BlockSize];
    }
    blocks[index][offset] = object;
    //
    // Adjust offset.
    //
    if (++offset == BlockSize) offset = 0;
}


void* GStack::pop (void)
{
    void* object = 0;

    if (nentries)
    {
        nentries--;

        if (offset == 0) offset = BlockSize;

        object = blocks[index][offset-1];

        if (--offset == 0)
            //
            // We can delete a block.
            //
            delete [] blocks[index--];
    }

    return object;
}


#ifdef TEST

#include <iostream.h>
#include <stdio.h>     /* for sprintf */
#include <new.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>


class StringStack : public GStack
{
  public:
    void push (char* str) { GStack::push(str); }
    char* pop (void) { return (char*)GStack::pop(); }
};


static void free_store_exhausted()
{
    cerr << "exiting, free store exhausted ...";
    exit(1);
}


static double elapsed_seconds(struct timeval& tp)
{
    static struct timezone tzp;
    static struct timeval  otp;
    otp = tp;
    gettimeofday(&tp, &tzp);
    return (tp.tv_sec - otp.tv_sec) + (tp.tv_usec - otp.tv_usec)/1000000.0;
}


static int equal(StringStack& s1, StringStack& s2)
{
    if (s1.entries() != s2.entries()) return 0;
    int rc = 1;
    for (int i = 0, max = (int)s1.entries(); i < max; i++)
    {
        if (strcmp(s1.pop(), s2.pop()))
        {
            rc = 0;
            break;
        }
    }
    return rc;
}
            

int main()
{
    set_new_handler(free_store_exhausted);

    StringStack s1;

    for (int i = 0; i < 33; i++)
    {
        char* name = new char[12];
        (void)sprintf(name, "%-10.10d", i);
        s1.push(name);
    }

    StringStack s2 = s1;
   
    cout << "test1: stacks are " << (equal(s1, s2) ? "equal\n" : "not equal\n");

    for (i = 0; i < 57; i++)
    {
        char* name = new char[12];
        (void)sprintf(name, "%-10.10d", i);
        s1.push(name);
    }

    s2 = s1;

    cout << "test2: stacks are " << (equal(s1, s2) ? "equal\n" : "not equal\n");

    const int count = 1000;

    struct timeval t;
    t.tv_sec  = 0;       // to keep CenterLine quiet
    t.tv_usec = 0;       // ditto

    elapsed_seconds(t);  // initialize timer

    srandom(13);         // Initialize the random number generator.

    for (i = 0; i < count; i++)
    {
        unsigned long number = (unsigned long)random();

        switch (number & 01)
        {
          case 0:
          {
              //
              // Push the number onto the stack.
              //
              char* name = new char[12];
              (void)sprintf(name, "%-10.10d", i);
              cout << "pushing " << name << "\n";
              s1.push(name);
              break;
          }
          case 1:
          {
              //
              // Pop the stack.
              //
              char* object = s1.pop();
              cout << "popping " << (object ==  0 ? "NULL" : object) << "\n";
              break;
          }
        }
    }

    cout << "count = " << count << ", seconds = " << elapsed_seconds(t) << endl;

    s2 = s1;

    cout << "test3: stacks are " << (equal(s1, s2) ? "equal\n" : "not equal\n");

    return 0;
}

#endif
