//
// Queue.C - queue for constant-sized objects. The objects must have a
//           working copy constructor as well as a default constructor.
//           The objects are stored in a dynamic array.  Note that all
//           builtin types satisfy these criterion.  If your compiler
//           supports templates compile with -DTEMPLATES.  Otherwise,
//           I use typedefs which must be hand set.
//
//
// $Id: Queue.C,v 1.8 1993/06/20 18:51:21 lijewski Exp $
//
// Copyright 1991, 1992, 1993 by Mike Lijewski.
//

#include <iostream.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "Queue.h"

#ifdef TEMPLATES
#ifdef TEST

void error (const char* msg) 
{
    cerr << msg << "\n";
}
     

extern "C" 
{
    double sin(double);
    double cos(double);
}


main (void)
{
    //
    // char queue test
    //
    char* alphabet = "abcdefghijklmnopqrstuvwxyz";
    
    Queue<char> qc(10);

    qc.stats();

    for (int i = 0; i < 10; i++)
    {
        for (char* p = alphabet; *p; p++)
            qc.insert(*p);
    }

    int entries = qc.entries();
    for (i = 0; i < entries/2; i++)
        cout << qc.get();
    cout << endl;

    qc.stats();

    Queue<char> qc2 = qc;

    qc2.stats();

    for (i = 0; i < 10; i++)
    {
        for (char* p = alphabet; *p; p++)
            qc2.insert(*p);
    }

    qc = qc2;

    while (!qc.isEmpty())
        cout << qc.get();
    cout << endl;

    qc2.stats();

    qc.stats();

    //
    // double queue test
    //

    Queue<double> qd(10);

    qd.stats();

    for (i = 0; i < 10; i++)
    {
        for (int j = 0; j < 20; j++)
            qd.insert(j * sin(j +.123));
    }

    entries = qd.entries();
    cout << "qd contains " << entries << " doubles\n";

    for (i = 0; i < entries/2; i++)
        cout << qd.get() << " ";
    cout << endl;

    qd.stats();

    Queue<double> qd2 = qd;

    qd2.stats();

    for (i = 0; i < 10; i++)
    {
        for (int j = 0; j < 20; j++)
            qd2.insert(j * cos(j +.123));
    }

    qd = qd2;

    while (!qd.isEmpty())
        cout << qd.get() << " ";
    cout << endl;

    qd2.stats();
    qd.stats();
}
#endif /*TEST&&TEMPLATES*/

#else /*!TEMPLATES*/


Queue::Queue (int capacity_estimate)
{
    head = tail = 0;
    rawdata = ::new T[(size = capacity_estimate)];
}


Queue::Queue (const Queue& q)
{
    head = q.head;
    tail = q.tail;
    size = q.size;
    rawdata = ::new T[size];
    memcpy(rawdata, q.data(), size * sizeof(T));
}


Queue& Queue::operator= (const Queue& q)
{
    if (this !=  &q)
    {
        delete [] rawdata;
        head = q.head;
        tail = q.tail;
        size = q.size;
        rawdata = ::new T[size];
        memcpy(rawdata, q.data(), size * sizeof(T));
    }
    return *this;
}


void Queue::insert (T object)
{
    if (tail == size)
    {
        if (head == 0)
        {
            T* nrawdata = ::new T[(size *= 2)];
            memcpy(nrawdata, rawdata, entries() * sizeof(T));
            delete [] rawdata;
            rawdata = nrawdata;
        }
        else
        {
#ifdef MEMMOVE
            memmove(rawdata, &rawdata[head], entries() * sizeof(T));
#else
            for (int i = 0; i < entries(); i++)
                rawdata[i] = rawdata[head + i];
#endif
            tail -= head;
            head = 0;
        }
    }
    rawdata[tail++] = object;
}


void Queue::error (void)
{
    ::error("nothing to get()\n");
}

#ifdef TEST

void error (const char* msg) 
{
    cerr << msg << "\n";
}

extern "C" 
{
    double sin(double);
    double cos(double);
}

main (void)
{
    //
    // char queue test
    //
    char* alphabet = "abcdefghijklmnopqrstuvwxyz";
    
    Queue q(10);

    for (int i = 0; i < 10; i++)
    {
        for (char* p = alphabet; *p; p++)
            q.insert(*p);
    }

    int entries = q.entries();
    for (i = 0; i < entries; i++)
        cout << q.get();
    cout << endl;

    Queue q2 = q;

    for (i = 0; i < 10; i++)
    {
        for (char* p = alphabet; *p; p++)
            q2.insert(*p);
    }

    q = q2;

    while (!q.isEmpty())
        cout << q.get();
    cout << endl;

#if 0
    //
    // double queue test
    //

    Queue q(10);

    for (int i = 0; i < 10; i++)
    {
        for (int j = 0; j < 20; j++)
            q.insert(j * sin(j +.123));
    }

    int entries = q.entries();
    cout << "q contains " << entries << " doubles\n";

    for (i = 0; i < entries; i++)
        cout << q.get() << " ";
    cout << endl;

    Queue q2 = q;

    for (i = 0; i < 10; i++)
    {
        for (int j = 0; j < 20; j++)
            q.insert(j * cos(j +.123));
    }

    q = q2;

    while (!q.isEmpty())
        cout << q.get() << " ";
    cout << endl;
#endif
}

#endif /*!TEMPLATES&&TEST*/

#endif /*!TEMPLATES*/
