//
// Queue.h - queue for constant-sized objects. The objects must have a
//           working copy constructor as well as a default constructor.
//           The objects are stored in a dynamic array.  Note that all
//           builtin types satisfy these criteria.  If your compiler
//           supports templates compile with -DTEMPLATES.  Otherwise,
//           I use typedefs which must be hand set.
//
// $Id: Queue.h,v 1.6 1993/06/20 18:51:21 lijewski Exp $
//
// Copyright 1991, 1992, 1993 by Mike Lijewski.
//


#ifndef QUEUE_H
#define QUEUE_H


#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include "dired.h"


#ifdef TEMPLATES
template<class T> class Queue
#else
//
// Change this typedef so that the type of object you want to
// store in the queue is referred to as TYPE.
//
typedef char T;

class Queue
#endif
{
  public:

    Queue (int initial_size);
    Queue(const Queue& rhs);
    
    ~Queue () { delete [] rawdata; }

    Queue& operator= (const Queue& rhs);

    T get () { if (isEmpty()) error(); return rawdata[head++]; }
    
    void insert (T object);

    int entries ()   const { return tail - head;    }
    const T* data () const { return &rawdata[head]; }
    int isEmpty ()   const { return entries() == 0; }

#ifdef TEST
    void stats (void);
#endif

  protected:

    void error (void);

    int size;
    int head;
    int tail;
    T* rawdata;
};


//
// If your compiler supports templates.
//
#ifdef TEMPLATES


template<class T> Queue<T>::Queue (int initial_size)
{
    head = tail = 0;
    size = initial_size;
    rawdata = ::new T[size];
}


template<class T> Queue<T>::Queue (const Queue<T>& q)
{
    head = 0;
    tail = q.entries();
    size = q.size;
    rawdata = ::new T[size];
    memcpy(rawdata, q.data(), q.entries() * sizeof(T));
}


template<class T> Queue<T>& Queue<T>::operator= (const Queue<T>& q)
{
    if (this != &q)
    {
        delete [] rawdata;
        head = 0;
        tail = q.entries();
        size = q.size;
        rawdata = ::new T[size];
        memcpy(rawdata, q.data(), q.entries() * sizeof(T));
    }
    return *this;
}


template<class T> void Queue<T>::insert (T object)
{
    if (tail == size)
    {
        if (head == 0)
        {
            size *= 2;
            T* nrawdata = ::new T[size];
            memcpy(nrawdata, rawdata, entries() * sizeof(T));
            delete [] rawdata;
            rawdata = nrawdata;
        }
        else
        {
#ifdef MEMMOVE
            memmove(rawdata, &rawdata[head], entries() * sizeof(T));
#else
            for (int i = 0; i < entries(); i++)
                rawdata[i] = rawdata[head + i];
#endif
            tail -= head;
            head = 0;
        }
    }
    rawdata[tail++] = object;
}


template<class T> void Queue<T>::error (void)
{
    ::error("nothing to get()\n");
}


#ifdef TEST
template<class T> void Queue<T>::stats (void)
{
    cerr << "entries = " << entries() << "\n"
         << "size    = " << size      << "\n"
         << "head    = " << head      << "\n"
         << "tail    = " << tail      << "\n";
}
#endif /*TEST*/


#endif /*TEMPLATES*/


#endif /*QUEUE_H*/
