//
// String.C - implementation of reference counted string class.
//
// $Id: String.C,v 1.7 1993/07/22 16:53:27 lijewski Exp $
//
// Copyright 1991, 1992, 1993 by Mike Lijewski
//

#include <new.h>
#include <stdio.h>
#include <string.h>

#include "String.h"


extern void error (const char *fmt, int index);  // For index range errors.

String::String (void) { p = new StringRep(); }

String::String (const char *s, size_t len) { p = new StringRep(s,len); }

String::String (char **s) { p = new StringRep(s, ::strlen(*s)); }

String::String (const char *s) { p = new StringRep(s); }

String::String (char** s, size_t slen) { p = new StringRep(s, slen); }


StringRep::StringRep (const char *s)
{
    len = ::strlen(s);
    rep = ::new char[len + 1];
    ::strcpy(rep, s);
    count = 1;
}


StringRep::StringRep (const char *s, size_t slen)
{
    rep = ::new char[slen + 1];
    ::strncpy(rep, s, slen);
    rep[slen] = 0;       // Force stringification.
    len = ::strlen(rep); // Force len == true length.
    count = 1;
}


String StringRep::operator+ (const String& s) const
{
    size_t slen  = s.length() + length();
    char *buf    = ::new char[slen + 1];
    ::strcpy(buf, rep);
    ::strcpy(buf + length(), s.p->rep);
    return String(&buf, slen);
}


//
// The definition of the head of the freelist that StringRep::operator new
// uses to dole out StringReps efficiently.
//

StringRep* StringRep::freeList;

void* StringRep::operator new (size_t size)
{
    if (size != sizeof(StringRep)) return ::new char[size];

    StringRep *s = freeList;

    if (s)
        freeList = s->next;
    else
    {
        //
        // We rely on a freestore exception catcher to catch
        // heap allocation failures.
        //
        StringRep *block = (StringRep*)::new char[chunksize*sizeof(StringRep)];

        for (int i = 0; i < chunksize - 1; i++)
            block[i].next = (StringRep *)&block[i + 1];
        block[chunksize - 1].next = 0;
        s = block;
        freeList = &block[1];
    }
    return s;
}


void StringRep::operator delete(void *object)
{
    StringRep *s = (StringRep *)object;
    s->next = freeList;
    freeList = s;
}


String::~String (void) { if (--p->count <= 0) delete p; }


String& String::operator= (const String& rhs)
{
    rhs.p->count++;
    if (--p->count <= 0) delete p;
    p = rhs.p;
    return *this;
}


void String::operator+= (const String& rhs)
{
    size_t slen = p->length() + rhs.length();
    char *buf   = ::new char[slen + 1];
    (void)strcpy(buf, p->rep);
    (void)strcpy(buf + p->length(), rhs.p->rep);
    if (p->count == 1)
    {
        delete [] p->rep;
        p->rep = buf;
        p->len = slen;
    }
    else
        operator=(String(&buf, slen));
}


void String::operator+= (const char *rhs)
{
    size_t slen = p->length() + ::strlen(rhs);
    char *buf = ::new char[slen + 1];
    ::strcpy(buf, p->rep);
    ::strcpy(buf + p->length(), rhs);
    if (p->count == 1)
    {
        delete [] p->rep;
        p->rep = buf;
        p->len = slen;
    }
    else
        operator=(String(&buf, slen));
}


void String::range_error (int index) const
{
    ::error("range error: %d out of bounds", index);
    exit(1);
}


//
// This operator WILL allow a constant String on the lhs of an = sign
// to have the char at index changed.  Use in this fashion at your own risk.
//
SBHelper String::operator[] (int index) const
{
    if (index < 0 || index >= length())
        range_error(index);
    String* str = (String*)this;
    return SBHelper(*str, index);
}


char SBHelper::operator= (char c)
{
    if (str.p->count == 1)
        //
        // Only one reference to our String.  Just assign the character to
        // the appropriate place.  Note that String::operator[] does the
        // range checking.
        //
        str.p->rep[index] = c;
    else
    {
        //
        // We have to uniquify our str.
        //
        str = String(str.p->rep);
        str.p->rep[index] = c;
    }
    return c;
}
