//
// String.h - a reference counted string class.
//
// $Id: String.h,v 1.5 1993/06/20 01:51:53 lijewski Exp $
//
// Copyright (c) 1991, 1992, 1993 by Mike Lijewski.
//

#ifndef DIRED__STRING_H
#define DIRED__STRING_H

#include <stddef.h>
#include <stdlib.h>
#include <string.h>


//
// A Simple reference counted string class.  It is implemented as an
// Envelope-Letter abstaction with String being the envelope and StringRep
// being the letter.
//

class String;
class SBHelper;

class StringRep
{
  public:

    friend class String;
    friend class SBHelper;

    StringRep ();
    StringRep (const char *s);
    StringRep (const char *s, size_t slen);
    StringRep (char** r, size_t slen);
    ~StringRep ();

    static StringRep *freeList;  // we manage our own storage
    enum { chunksize = 50 };     // # of StringReps to allocate at a time
    void *operator new (size_t size);
    void operator delete (void *object);

    int operator!= (const char *rhs)      const;
    int operator== (const char *rhs)      const;
    int operator!= (const StringRep& rhs) const;
    int operator== (const StringRep& rhs) const;

    String operator+ (const String& s) const;

    size_t length () const;

  private:

    //
    // Disable these two methods
    //
    StringRep (const StringRep&);
    StringRep& operator= (const StringRep &);

    union {
        char* rep;
        StringRep* next;
    };
    size_t len;
    int count;
};


class String
{
  public:

    friend class StringRep;
    friend class SBHelper;

    String ();
    String (const String& s);
    String (const char *s);
    String (const char *s, size_t len);
    String (char **s);
    String (char** s, size_t slen);
    ~String ();

    String& operator= (const String& rhs);

    int operator== (const char *rhs)   const;
    int operator== (const String& rhs) const;
    int operator!= (const char *rhs)   const;
    int operator!= (const String& rhs) const;

    String operator+ (const String &rhs) const;
    friend String operator+ (const char *lhs, const String& rhs);

    void operator+= (const String &rhs);
    void operator+= (const char *rhs);

    operator const char * () const;
    SBHelper operator[] (int index) const;
    size_t length () const;

  private:

    void range_error (int index) const;

    StringRep* p;
};


//
// This class is a helper class used by String::operator[] to distinguish
// between applications of operator[] on the lhs and rhs of "=" signs.
//

class SBHelper
{
  public:

    SBHelper (String &s, int i);
    SBHelper (const SBHelper& s);
    char operator= (char c);
    operator int () const;

  private:

    void operator= (const SBHelper&);  // disallow this method

    String& str;
    int index;
};


//
// StringRep inlines.
//

inline StringRep::StringRep () { rep = ::new char[1]; len=0; *rep=0; count=1; }

inline StringRep::StringRep (char** r, size_t slen) {rep=*r;len=slen;count=1;}

inline StringRep::~StringRep () { delete [] rep; }

inline int StringRep::operator!= (const char *rhs) const
{
    return strcmp(rep, rhs);
}

inline int StringRep::operator== (const char *rhs) const
{
    return strcmp(rep, rhs) == 0;
}

inline int StringRep::operator!= (const StringRep& rhs) const
{
    return strcmp(rep, rhs.rep);
}

inline int StringRep::operator== (const StringRep& rhs) const
{
    return strcmp(rep, rhs.rep) == 0;
}

inline size_t StringRep::length () const { return len; }


//
// String inlines.
//

inline String::String (const String& s) { p = s.p; p->count++; }

inline int String::operator== (const char *rhs) const { return *p == rhs; }

inline int String::operator== (const String& rhs) const {return *p==*(rhs.p);}

inline int String::operator!= (const char *rhs) const { return *p != rhs; }

inline int String::operator!= (const String& rhs) const {return *p!=*(rhs.p);}

inline String String::operator+ (const String &rhs) const { return *p + rhs; }

inline String operator+ (const char *lhs, const String& rhs)
{
    return rhs + String(lhs);
}

inline String::operator const char * () const { return p->rep; }

inline size_t String::length () const { return p->len; }


//
// SBHelper inlines.
//

inline SBHelper::SBHelper (String& s, int i) : str(s), index(i) { };

inline SBHelper::SBHelper (const SBHelper& s) : str(s.str), index(s.index) { };

inline SBHelper::operator int () const { return str.p->rep[index]; }

#endif /* DIRED_STRING_H */
