//
// classes.h - the declarations of the classes used in keeping track
//             of the lines in a directory and their placement on the screen.
//
// $Id: classes.h,v 1.5 1993/06/20 01:51:53 lijewski Exp $
//
// Copyright (c) 1991, 1992, 1993 by Mike Lijewski.
//

#ifndef DIRED_CLASSES_H
#define DIRED_CLASSES_H

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "GStack.h"
#include "String.h"


//
// A class which contains one line of the long listing of a directory.
//

class DirLine
{
  public:

    friend class DirList;

    DirLine (char **line);
    ~DirLine ();

    static DirLine *freeList;  // we manage our own storage for DirLines
    enum { chunksize = 50 };   // size blocks of DirLines we allocate
    void *operator new(size_t size);
    void operator delete(void *object);

    const char* line () const;
    int length () const;

    DirLine* next () const;
    DirLine* prev () const;

    void update (char **newline);

  private:

    //
    // Disallow these operations by not providing definitions.
    // Also keep compiler from generating default versions of these.
    //
    DirLine ();
    DirLine (const DirLine &);
    DirLine& operator= (const DirLine &); 

    String   _line;
    DirLine* _next;
    DirLine* _prev;
};


//
// A class which manages a doubly-linked list of DirLines.
// It also maintains our current notion of what is and isn't
// visible in the window.
//

class DirList
{
  public:

    DirList (char *);
    ~DirList ();

    DirLine* head () const;
    DirLine* tail () const;
    DirLine* firstLine () const;
    DirLine* lastLine () const;
    DirLine* currLine () const;

    DirList* next () const;
    DirList* prev () const;

    int savedXPos () const;
    int savedYPos () const;

    void setFirst (DirLine *e);
    void setLast (DirLine *e);
    void setCurrLine (DirLine *ln);

    void setNext (DirList *l);
    void setPrev (DirList *l);

    const char* name () const;
    int nelems () const;
    void saveYXPos (int y, int x);

    int atBegOfList () const;
    int atEndOfList () const;

    int atWindowTop () const;
    int atWindowBot () const;

    DirLine* insert (char **);
    void add(DirLine *);
    void deleteLine ();

  private:

    //
    // Disallow these operations by not providing definitions.
    // Also keep compiler from generating default versions of these.
    //
    DirList ();
    DirList (const DirList &);
    DirList& operator= (const DirList &);

    DirLine* _head;
    DirLine* _tail;
    char*    _name;        // full pathname of the directory
    int      _nelems;
    short    _saved_x;     // saved x cursor position
    short    _saved_y;     // saved y cursor position
    DirLine* _firstLine;   // first viewable dirLine in curses window
    DirLine* _lastLine;    // last  viewable dirLine in curses window
    DirLine* _currLine;    // line cursor is on in curses window
    DirList* _next;
    DirList* _prev;
};


//
// A class which manages a stack of DirLists.
//

class DirStack : public GStack
{
  public:

    void push (DirList *list);
    DirList* pop ();

    unsigned long nelems () const;
    DirList* top () const;
};


//
// DirLine inlines.
//

inline DirLine::DirLine (char **line) : _line(line) { _next = _prev = 0; }

inline DirLine::~DirLine () { }

inline const char* DirLine::line () const { return _line; }

inline int DirLine::length () const { return _line.length(); }

inline DirLine* DirLine::next () const { return _next; }

inline DirLine* DirLine::prev () const { return _prev; }

inline void DirLine::update (char **newline) { _line = String(newline); }


//
// DirList inlines.
//

inline DirLine* DirList::head () const { return _head; }

inline DirLine* DirList::tail () const { return _tail; }

inline DirLine* DirList::firstLine () const { return _firstLine; }

inline DirLine* DirList::lastLine () const { return _lastLine; }

inline DirLine* DirList::currLine () const { return _currLine; }

inline DirList* DirList::next () const { return _next; }

inline DirList* DirList::prev () const { return _prev; }

inline int DirList::savedXPos ()  const { return _saved_x; }

inline int DirList::savedYPos ()  const { return _saved_y; }

inline void DirList::setFirst (DirLine *e) { _firstLine = e; }

inline void DirList::setLast (DirLine *e) { _lastLine  = e; }

inline void DirList::setCurrLine (DirLine *ln) { _currLine = ln; }

inline void DirList::setNext (DirList *l) { _next = l; }

inline void DirList::setPrev (DirList *l) { _prev = l; }

inline const char* DirList::name () const { return _name; }

inline int DirList::nelems () const { return _nelems; }

inline void DirList::saveYXPos (int y, int x)
{
    _saved_x = (short)x;
    _saved_y = (short)y;
}

inline int DirList::atBegOfList () const { return _currLine == _head; }

inline int DirList::atEndOfList () const { return _currLine == _tail; }

inline int DirList::atWindowTop () const { return _currLine == _firstLine; }

inline int DirList::atWindowBot () const { return _currLine == _lastLine;  }


//
// DirStack inlines.
//

inline DirList* DirStack::top () const { return (DirList*) GStack::top(); }

inline void DirStack::push (DirList *list) { GStack::push(list); }

inline DirList* DirStack::pop () { return (DirList*) GStack::pop(); }

inline unsigned long DirStack::nelems () const { return GStack::entries(); }


#endif /* DIRED_CLASSES_H */
