//
// command1.C - some of the commands called from the main command loop.
//              Command1.C and command2.C are concatenated during the
//              make into commands.C, which consists of the main command
//              loop and all commands called from within that loop.
//
// $Id: command1.C,v 1.15 1993/09/27 19:40:09 lijewski Exp $
//
// Copyright (c) 1991, 1992, 1993 by Mike Lijewski.
//


#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef sgi
#include <sys/sysmacros.h>  /* to get minor() and major() macros */
#endif

#ifdef apollo	/* no prototypes defined for these... */
extern "C"
{
    int	read    (int, const char *, int);
    int	link    (const char *, const char *);
    int	symlink (const char *, const char *);
    int	unlink  (const char *);
    int	rmdir   (const char *);
    int	chdir   (const char *);
}
#else
#include <unistd.h>
#endif

#include "String.h"
#include "commands.h"
#include "dired.h"
#include "display.h"
#include "keys.h"
#include "version.h"


//
// These functions aren't prototyped under ObjectCenter 2.0 on my Sun.
//
#if defined(sun) && defined(__CLCC__)
extern "C"
{
	int symlink (const char*, const char*);
	int rename  (const char*, const char*);
}
#endif


//
// Scroll the listing up one line. We only call this routine when
// we KNOW that there is at least one line below the window which
// can be scrolled into it and the cursor is on the last line of the screen.
//

static void scroll_up_one_line (DirList *dl)
{
    dl->setFirst(dl->firstLine()->next());
    dl->setLast(dl->lastLine()->next());
    dl->setCurrLine(dl->lastLine());

    if (CS)
    {
        scroll_listing_up_one();
        display_string(dl->currLine()->line(), dl->currLine()->length());
    }
    else if (DL || SF)
    {
        clear_modeline();
        scroll_screen_up_one();
        update_modeline();
        move_cursor(rows()-3, 0);
        display_string(dl->currLine()->line(), dl->currLine()->length());
    }
    else
        redisplay();

    dl->saveYXPos(rows()-3, goal_column(dl));
    move_cursor(rows()-3, dl->savedXPos());
}


//
// Move cursor down one line.
//

void cursor_down_one_line (DirList* dl)
{
    if (dl->atEndOfList()) return;
    
    if (dl->currLine()->length() > columns()) 
        rightshift_current_line(dl);
    
    if (dl->savedYPos() < rows() - 3)
    {
        //
        // There are still more lines below us in the window
        // so we just move the cursor down one line.
        //
        dl->setCurrLine(dl->currLine()->next());
        int x = goal_column(dl);
        if (x == dl->savedXPos())
            cursor_down();
        else
            move_cursor(dl->savedYPos() + 1, x);
        dl->saveYXPos(dl->savedYPos() + 1, x);
    }
    else
        //
        // We are on the last line on the screen and there
        // are more lines to display.  Scroll up one line
        // and leave the cursor on the next logical line.
        //
        scroll_up_one_line(dl);
    
    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    
    synch_display();
}


//
// Scroll the listing down one line. We only call this routine when we KNOW
// that the head of the listing is not visible and the cursor is on the
// first line in the window.
//

static void scroll_down_one_line (DirList *dl)
{
    if (lines_displayed(dl) == rows() - 2)
        //
        // Must update lastLine.
        // We previously had a screenfull of lines.
        //
        dl->setLast(dl->lastLine()->prev());

    dl->setFirst(dl->firstLine()->prev());
    dl->setCurrLine(dl->firstLine());

    if (CS)
    {
        scroll_listing_down_one();
        display_string(dl->currLine()->line(), dl->currLine()->length());
    }
    else if (AL || SR)
    {
        clear_modeline();
        scroll_screen_down_one();
        update_modeline();
        cursor_home();
        display_string(dl->currLine()->line(), dl->currLine()->length());
    }
    else
        redisplay();

    dl->saveYXPos(0, goal_column(dl));
    move_cursor(0, dl->savedXPos());
}


//
// Move cursor up one line.
//

void cursor_up_one_line (DirList* dl)
{
    if (dl->atBegOfList()) return;
    
    if (dl->currLine()->length() > columns())
        rightshift_current_line(dl);
    
    if (dl->savedYPos() != 0)
    {
        //
        // We are not at the top of the window so can move up.
        //
        dl->setCurrLine(dl->currLine()->prev());
        int x = goal_column(dl);
        if (x == dl->savedXPos() && UP)
            cursor_up();
        else
            move_cursor(dl->savedYPos() - 1, x);
        dl->saveYXPos(dl->savedYPos() - 1, x);
    }
    else
        //
        // We are on the first line of the window and there are
        // lines preceding us in the directory listing.
        //
        scroll_down_one_line(dl);
    
    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    
    synch_display();
}


//
// Scroll listing up one full window, leaving one line of overlap.
//

void scroll_up_full_window (DirList *dl)
{
    if (dl->lastLine() == dl->tail())
        return;

    DirLine *ln = dl->lastLine();
    dl->setFirst(ln);
    dl->setCurrLine(ln);

    clear_display_area();
    
    for (int i = 0; i < rows() - 2 && ln; i++, ln = ln->next())
        display_string(ln->line(), ln->length());
    
    if (ln)
        dl->setLast(ln->prev());
    else
        dl->setLast(dl->tail());

    dl->saveYXPos(0, goal_column(dl));

    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(0, dl->savedXPos());
    
    synch_display();
}


//
// Try to scroll listing down one full window, with one line of overlap.
// This routine is only called when we KNOW that there is at least one
// line "above" the current listing. Only change the current line if it
// flows off the "bottom" of the screen.
//

void scroll_down_full_window (DirList *dl)
{
    if (dl->firstLine() == dl->head())
        return;

    DirLine *ln = dl->firstLine();
    for (int y = 0; y < rows() - 3 && ln != dl->head(); y++, ln = ln->prev())
        ;

    //
    // y == # of lines preceding firstLine to add to screen.
    //
    
    dl->setFirst(ln);

    clear_display_area();
    
    for (int j = 0; j < rows()-2 && ln; j++, ln = ln->next())
        display_string(ln->line(), ln->length());

    if (ln)
        dl->setLast(ln->prev());
    
    if (dl->savedYPos()+y >= rows()-2)
    {
        dl->setCurrLine(dl->lastLine());
        dl->saveYXPos(rows()-3, goal_column(dl));
    }
    else
        dl->saveYXPos(dl->savedYPos()+y, dl->savedXPos());

    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(dl->savedYPos(), dl->savedXPos());
    
    synch_display();
}


//
// Scroll listing up half a window. We try to leave the cursor
// on the file it was on previously, otherwise it is left on the
// first file in the screen.
//

void scroll_up_half_window (DirList *dl)
{
    if (dl->lastLine() == dl->tail())
        return;

    if (dl->currLine()->length() > columns())
        rightshift_current_line(dl);

    DirLine *ln = dl->firstLine();
    for (int i = 0; i < (rows() - 2)/2; i++, ln = ln->next())
        ;
    dl->setFirst(ln);
    
    if (CS || DL || SF || DLN)
    {
        if (CS)
            scroll_listing_up_N((rows()-2)/2);
        else
        {
            clear_modeline();
            scroll_screen_up_N((rows()-2)/2);
            update_modeline();
        }
        move_cursor(rows() - 2 -((rows()-2)/2), 0);
        ln = dl->lastLine()->next();
        for (i = 0; i < (rows() - 2)/2 && ln; i++, ln = ln->next())
            display_string(ln->line(), ln->length());

        if (ln)
            dl->setLast(ln->prev());
        else
            dl->setLast(dl->tail());
    }
    else
    {
        clear_display_area();
        
        for (i = 0; i < rows() - 2 && ln->next(); i++, ln = ln->next())
            display_string(ln->line(), ln->length());
        
        if (i != rows()-2)
        {
            //
            // We hit last line before outputing all that we could.
            // Must output lastLine == tail.
            //
            display_string(ln->line(), ln->length());
            dl->setLast(ln);
        }
        else
            dl->setLast(ln->prev());
    }

    int pos = dl->savedYPos() - (rows()-2)/2;
    if (pos < 0)
    {
        pos = 0;
        dl->setCurrLine(dl->firstLine());
    }
    
    dl->saveYXPos(pos, goal_column(dl));

    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(pos, dl->savedXPos());
    
    synch_display();
}


//
// Try to scroll listing down half a window. If freshen is true, which
// is the default, the screen is refreshed.  It is important to note
// that we may not be able to scroll down a complete half window, since we
// always anchor the head of the listing to the first line in the screen.
//

void scroll_down_half_window (DirList *dl, int freshen)
{
    if (dl->firstLine() == dl->head())
        return;

    if (dl->firstLine() != dl->head())
    {
        //
        // We can scroll down.  Try to leave the cursor on the file
        // it started out on.  Otherwise, leave it on the
        // (rows-2)/2 line, which was the previous firstLine.
        //
        DirLine *ln = dl->firstLine();
        for (int i = 0; i < (rows()-2)/2 && ln->prev(); i++, ln = ln->prev())
            ;
        dl->setFirst(ln);

        if (dl->currLine()->length() > columns())
            rightshift_current_line(dl);

        if (CS || AL || ALN || SR)
        {
            if (CS)
                scroll_listing_down_N(i);
            else
            {
                clear_modeline();
                scroll_screen_down_N(i);
                update_modeline();
                clear_message_line();
            }
            cursor_home();
            for (int j = 0; j < i; j++, ln = ln->next())
                display_string(ln->line(), ln->length());
            ln = dl->firstLine();
            for (int i = 0; i < rows()-2 && ln->next(); i++, ln = ln->next())
                ;
            dl->setLast(ln);
        }
        else
        {
            clear_display_area();
            for (int i = 0; i < rows()-2 && ln->next(); i++, ln = ln->next())
                display_string(ln->line(), ln->length());
            
            if (i != rows() - 2)
            {
                //
                // We hit last line before outputing all that we could.
                // Must output lastLine == tail.
                //
                display_string(ln->line(), ln->length());
                dl->setLast(ln);
                i++;  // So we know how many lines have been written;
            }
            else
                dl->setLast(ln->prev());
        }

        int pos = i + dl->savedYPos();
        if (pos > rows() - 3)
        {
            pos = rows() - 3;
            dl->setCurrLine(dl->lastLine());
        }

        dl->saveYXPos(pos, goal_column(dl));

        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(pos, dl->savedXPos());

        if (freshen)
            synch_display();
    }
}


//
// Position cursor on first line in listing.
//

void goto_first (DirList *dl)
{
    if (dl->atBegOfList())
        return;

    if (dl->head() != dl->firstLine())
        initial_listing(dl);
    else
    {
        if (dl->currLine()->length() > columns())
            rightshift_current_line(dl);
        dl->setCurrLine(dl->head());
    }

    dl->saveYXPos(0, goal_column(dl));
    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(0, dl->savedXPos());

    synch_display();
}


//
// Position cursor on last file in listing.
//

void goto_last (DirList *dl)
{
    if (dl->atEndOfList())
        return;

    if (dl->currLine()->length() > columns())
        rightshift_current_line(dl);

    dl->setCurrLine(dl->tail());

    if (dl->tail() == dl->lastLine())
    {
        //
        // Only need to reposition the cursor.
        //
        dl->saveYXPos(lines_displayed(dl) - 1, goal_column(dl));
        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    else
    {
        //
        // Redisplay end of listing & update our view.
        //
        DirLine *ln = dl->tail();
        dl->setLast(ln);

        clear_display_area();
        
        for (int i = 0; i < rows() - 2; i++, ln = ln->prev())
        {
            move_cursor(rows() - 3 - i, 0);
            display_string(ln->line(), ln->length());
        }
        dl->setFirst(ln->next());
        dl->saveYXPos(rows() - 3,goal_column(dl));

        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(rows() -3 , dl->savedXPos());
    }

    synch_display();
}


//
// Prompt for a directory to edit and edit it, provided it is a valid
// directory. If the first character of the directory is tilde, the
// tilde is expanded to the user's home directory.
//

void edit_prompted_for_directory (DirList *dl, const KeyMap& keymap)
{
#ifdef COMPLETION
    const char *dir = prompt("Directory to edit: ", 1); // prompt w/ completion
#else
    const char *dir = prompt("Directory to edit: ");
#endif

    if (dir == 0)
    {
        //
        // Command was aborted.
        //
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    dir = expand_tilde(dir);

    if (!is_directory(dir))
    {
        message("`%' is not a valid directory name", dir);
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    if (!read_and_exec_perm(dir))
    {
        message("need read & exec permission to edit `%'", dir);
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    dired(dir, keymap);
}


//
// Rereads the listing line for the current line in the DirList,
// which refers to file, and then updates the listing line.
//

static void reread_listing_line (DirList *dl, const char *file)
{
    String cmd(ls_cmd[the_sort_order()]);
    cmd += "'";
    cmd += file;
    cmd += "'";

    FILE *fp = popen(cmd, "r");
    if (fp == 0)
        error("File %s, line %d: popen(%s) failed",
              __FILE__, __LINE__, (const char *)cmd);

    char *new_line = fgetline(fp);
    if (fp == 0)
        error("File %s, line %d: fgetline() failed", __FILE__, __LINE__);
    (void)pclose(fp);

    dl->currLine()->update(&new_line);
}


//
// Edit the current file in the DirList.
//

void edit_current_file (DirList *dl, const KeyMap& keymap)
{
    const char *file = get_file_name(dl);
    
    if (is_directory(file))
    {
        if (strcmp(file, ".") == 0 || (strcmp(file, "..") == 0 &&
                                       strcmp(dl->name(), "/") == 0))
            return;

        if (!read_and_exec_perm(file))
        {
            message("need read & exec permissions to edit `%'", file);
            move_cursor(dl->savedYPos(), dl->savedXPos());
            synch_display();
            return;
        }

        dired(file, keymap);
    }
    else if (is_regular_file(file))
    {
        char *editor = getenv("EDITOR");
        if (editor == 0)
            editor = "vi";

        String cmd(editor);
        cmd += " '";
        cmd += file;
        cmd += "'";

        exec_with_system(cmd, 0);

        //
        // Re-read the listing line for this file
        // and insert into the directory listing.
        // This way, changes in file characteristics are
        // reflected in the soon-to-be updated listing.
        //
        reread_listing_line(dl, file);
        redisplay();
    }
    else
    {
        message("can only edit regular files and directories");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
    }
}


//
// Ask user to type any key to continue.
//

inline void type_any_key_to_continue (void)
{
    eat_a_character("Press Any Key to Continue");
}


//
// Attempt to page the current file in the DirList.
// If we have a directory, we edit it.
//

void page_current_file (DirList *dl, const KeyMap& keymap)
{
    const char *file = get_file_name(dl);
    if (is_directory(file))
    {
        if (strcmp(file, ".") == 0 || (strcmp(file, "..") == 0 &&
                                       strcmp(dl->name(), "/") == 0))
            return;

        if (!read_and_exec_perm(file))
        {
            message("need read & exec permissions to edit `%'", file);
            move_cursor(dl->savedYPos(), dl->savedXPos());
            synch_display();
            return;
        }

        dired(file, keymap);
    }
    else if (is_regular_file(file))
    {
        char *pager = getenv("PAGER");
        if (pager == 0)
            pager = "more";

        String cmd(pager);
        cmd += " '";
        cmd += file;
        cmd += "'";

        exec_with_system(cmd);

        if (the_sort_order() == ACCESS_TIME ||
            the_sort_order() == INODE_CHANGE_TIME)
            //
            // Re-read the listing line for this file
            // and insert into the directory listing.
            // This way, changes in file characteristics are
            // reflected in the soon-to-be updated listing.
            // Since we have only paged the file, the listing should only
            // change, if we are interested in the access or inode-
            // modification times.
            //
            reread_listing_line(dl, file); 
        redisplay();
    }
    else
    {
        message("can only page through regular files and directories");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
    }
}


//
// Inserts the current line into the DirList after line y. Only called
// if y != rows-3.  That is, this routine is for the case when we do
// not need to scroll the screen to place the line into its logical
// place on the screen.
//

static void insert_line (DirList *dl, int y )
{
    if (CS)
    {
        insert_listing_line(y + 1);
        display_string(dl->currLine()->line(), dl->currLine()->length());
    }
    else if(AL)
    {
        clear_modeline();
        insert_blank_line(y + 1);
        display_string(dl->currLine()->line(), dl->currLine()->length());
        update_modeline();
    }
    else
        redisplay();
}


//
// Inserts the listing line for dest into the DirList after line y
// in the current window.
//

static void insert_into_dirlist (DirList *dl, const char *dest, int y)
{
    char *pos = strrchr(dest, '/');
    //
    // pos is non-zero in those cases where dest contains
    // `..' trickery.  Say, ../this-dir/new-file.
    //
    String command(ls_cmd[the_sort_order()]);
    command += "'";
    command += (pos ? pos + 1 : dest);
    command += "'";

    FILE *fp = popen(command, "r");
    if (fp == 0)
        error("File %s, line %d: popen(%s) failed",
              __FILE__, __LINE__, (const char *)command);

    char *new_line = fgetline(fp);
    if (fp == 0)
        error("File %s, line %d: fgetline() failed", __FILE__, __LINE__);
    (void)pclose(fp);
    
    int nlines = lines_displayed(dl);
    if (dl->currLine()->length() > columns())
        rightshift_current_line(dl);
    dl->setCurrLine(dl->insert(&new_line));
    
    if (nlines == rows() - 2)
    {
        if (y == rows() - 3)
        {
            //
            // We must scroll listing up.
            //
            dl->setFirst(dl->firstLine()->next());
            dl->setLast(dl->currLine());

            if (CS)
            {
                scroll_listing_up_one();
                display_string(dl->currLine()->line(), dl->currLine()->length());
            }
            else if (DL || SF)
            {
                clear_modeline();
                scroll_screen_up_one();
                update_modeline();
                move_cursor(rows()-3, 0);
                display_string(dl->currLine()->line(), dl->currLine()->length());
            }
            else
                redisplay();

            dl->saveYXPos(y, goal_column(dl));
            move_cursor(y, dl->savedXPos());
        }
        else
        {
            //
            // Just insert line.
            //
            dl->setLast(dl->lastLine()->prev());
            insert_line(dl, y);
            dl->saveYXPos(y+1, goal_column(dl));
            move_cursor(y+1, dl->savedXPos());
        }
    }
    else
    {
        insert_line(dl, y);
        dl->saveYXPos(y + 1, goal_column(dl));
        move_cursor(y + 1, dl->savedXPos());
        if (nlines == y + 1)
            //
            // The new line becomes the new lastLine.
            //
            dl->setLast(dl->currLine());
    }

    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
}


//
// Returns true if a y or Y is typed in response to the msg.
// Deals with SIGWINCH and SIGTSTP. Does not synch the display.
//

static int yes_or_no (const char *msg)
{
    message(msg);
    char c;
    while (1)
        if (read(0, &c, 1) < 0 // assume fails only when errno == EINTR
#ifdef SIGWINCH
            || win_size_changed
#endif
            )
        {
#ifdef SIGWINCH
            if (win_size_changed)
            {
                win_size_changed = 0;
                adjust_window();
                redisplay();
            }
#endif
            message(msg);
        }
        else
        {
            clear_message_line();
            return c == 'y' || c == 'Y';
        }

}


//
// Search through the line list, starting with the given DirLine,
// for a line whose filename contains a match for str.  The
// matching is strictly a string match using strstr, not a
// regex match.  Returns the first matching line found, or 0
// if none found.  pos is a pointer to an int containing
// the position in the window of ln.  On exit, it contains the
// position of the matching line, relative to the starting position.
// If forward is nonzero (the default), we search forward,
// otherwise we search backwards.  If str is NULL, we redo the
// previous search.
//

static DirLine *search (DirLine *ln, const char *str, int *pos, char forward=1)
{
    static String saved_str;
    const char *name;
    
    if (*str)
        //
        // Save copy of str so we can redo searches.
        //
        saved_str = str;
    else if (saved_str != "")
        //
        // We have a previous search string.
        //
        str = saved_str;
    else
        return 0;

    while (ln)
    {
#ifndef NO_SYMLINKS
        if ((name = BMstrstr(ln->line(), " -> ")) != 0)
            //
            // We have a symbolic link.
            //
            --name;
        else
#endif
            name = ln->line() + ln->length();

        while (!isspace(*name)) --name;
        if (BMstrstr(name, str))
            return ln;

        if (forward)
        {
            ln = ln->next();
            (*pos)++;
        }
        else
        {
            ln = ln->prev();
            (*pos)--;
        }
    }

    return 0;
}


//
// Returns true if the file file, which is relative to the DirList dl,
// is actually in the DirList's directory. Returns false on any error.
//

static int in_same_directory (DirList *dl, const char *file)
{
    const char *last = strrchr(file, '/');

    //
    // If there are no slashes in the filename, it has to be in the
    // same directory.
    //
    if (last == 0)
        return 1;

    if (last == file)
        last += 1;  // Watch out for root.
    String dir(file);
    dir[last - file] = 0;
    struct stat buf1, buf2;

#ifdef NO_SYMLINKS
    if (stat((const char*)dir, &buf1) < 0 || stat(dl->name(), &buf2) < 0)
#else
    if (lstat((const char*)dir, &buf1) < 0 || lstat(dl->name(), &buf2) < 0)
#endif
        return 0;

    //
    // Otherwise, if its parent directory and the starting directory have
    // the same inode number and device number, the file is in the same
    // directory.
    //
    return buf1.st_ino == buf2.st_ino               &&
           major(buf1.st_dev) == major(buf2.st_dev) &&
           minor(buf1.st_dev) == minor(buf2.st_dev);
}


//
// Copy current file to destination. Update window appropriately.
//

static void copy_file (DirList *dl, const char *src, const char *dest)
{
    const char *cp = "cp";
    const char *args[4];

    args[0] = cp;
    args[1] = src;
    args[2] = dest;
    args[3] = 0;

    if (is_regular_file(dest))
    {
        String msg("overwrite '");
        msg += dest;
        msg += "'? (y|n) ";

        if (yes_or_no(msg))
            if (execute(cp, args))
            {
                message("`copy' was successful");

                if (in_same_directory(dl, dest))
                {
                    //
                    // dest is in our current directory.
                    // It may contain /s due to .. or ~ funny business.
                    //
                    int pos = 0;
                    const char *slash = strrchr(dest, '/');
                    if (slash)
                        dest = slash + 1;
                    DirLine *found = search(dl->firstLine(), dest, &pos);
                    if (found && pos < rows() - 2)
                    {
                        //
                        // And it is also in our current window.
                        //
                        String cmd(ls_cmd[the_sort_order()]);
                        cmd += "'";
                        cmd += dest;
                        cmd += "'";

                        FILE *fp = popen(cmd, "r");
                        if (fp == 0)
                            error("File %s, line %d: popen(%s) failed",
                                  __FILE__, __LINE__, (const char *)cmd);

                        char *new_line = fgetline(fp);
                        if (fp == 0)
                            error("File %s, line %d: fgetline() failed",
                                  __FILE__, __LINE__);
                        (void)pclose(fp);
            
                        update_screen_line(found->line(), new_line, pos);
                        found->update(&new_line);
                    }
                }
            }
            else
                message("`copy' failed");

        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    else if (is_directory(dest))
    {
        String file(dest);
        if (dest[strlen(dest) - 1] != '/')
            file += "/";
        file += src;

        if (is_regular_file(file))
        {
            String msg("overwrite `");
            msg += file;
            msg += "'? (y|n) ";

            if (yes_or_no(msg))
                if (execute(cp, args))
                    message("`copy' was successful");
                else
                    message("`copy' failed");
        }
        else
        {
            //
            // Just do the cp.
            //
            if (execute(cp, args))
                message("`copy' was successful");
            else
                message("`copy' failed");
        }

        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    else
    {
        //
        // Dest does not already exist.
        //
        if (execute(cp, args))
        {
            message("`copy' was successful");
            //
            // Is dest in our current directory?
            //
            if (in_same_directory(dl, dest))
                insert_into_dirlist(dl, dest, dl->savedYPos());
            else
                move_cursor(dl->savedYPos(), dl->savedXPos());
        }
        else
        {
            message("`copy' failed");
            move_cursor(dl->savedYPos(), dl->savedXPos());
        }
    }
}


//
// Attempt to copy current file to another.
//

void copy_current_file (DirList *dl)
{
    const char *src = get_file_name(dl);

    if (strcmp(src, ".")  == 0 || strcmp(src, "..") == 0)
    {
        message("`copying' of `.' and `..' is not allowed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
#ifdef COMPLETION
    const char *dest = prompt("Copy to: ", 1);  // Prompt with completion.
#else
    const char *dest = prompt("Copy to: ");
#endif
    
    if (dest == 0)
    {
        //
        // Command was aborted.
        //
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*dest == 0)
    {
        message("not a valid file name");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    dest = expand_tilde(dest);

    copy_file(dl, src, dest);

    synch_display();
}


//
// Attempt to link file to dest.
//

static void link_file (DirList *dl, const char *file, const char *dest)
{
    if (!link(file, dest))
    {
        message("`link' was successful");
        if (in_same_directory(dl, dest))
            insert_into_dirlist(dl, dest, dl->savedYPos());
        else
            move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    else
    {
        message("`link' failed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
}


//
// Attempt to make a link to the current file.
//

void link_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    const char *link = prompt("Link to: ");

    if (link == 0)
    {
        //
        // Command was aborted.
        //
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*link == 0)
    {
        message("not a valid file name");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    link = expand_tilde(link);

    link_file(dl, file, link);

    synch_display();
}


#ifndef NO_SYMLINKS
//
// Symbolically link file to dest.
//

static void symlink_file (DirList *dl, const char *file, const char *dest)
{
    if (in_same_directory(dl, dest))
    {
        //
        // Don't need to expand pathname of `file'.
        //
        if (!symlink(file, dest))
        {
            message("`symlink' was successful");
            insert_into_dirlist(dl, dest, dl->savedYPos());
        }
        else
        {
            message("`symlink' failed");
            move_cursor(dl->savedYPos(), dl->savedXPos());
        }
    }
    else
    {
        //
        // Use full pathname of `file'.
        //
        String pathname(dl->name());
        if (pathname != "/")
            pathname += "/";
        pathname += file;

        if (!symlink((const char*)pathname, dest))
        {
            message("`symlink' was successful");
            move_cursor(dl->savedYPos(), dl->savedXPos());
        }
        else
        {
            message("`symlink' failed");
            move_cursor(dl->savedYPos(), dl->savedXPos());
        }
    }
}


//
// Attempt to create a symbolic link to the current file.
//

void symlink_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    const char *link = prompt("Name of symbolic link: ");
    
    if (link == 0)
    {
        //
        // Command was aborted.
        //
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*link == 0)
    {
        message("not a valid file name");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    link = expand_tilde(link);

    symlink_file(dl, file, link);

    synch_display();
}
#endif


//
// Delete the current line in the DirList and update both the screen
// and data structures appropriately. `y' is the position in the window
// of the current line.
//

static void remove_listing_line (DirList *dl, int y)
{
    if (dl->lastLine() != dl->tail())
    {
        //
        // Last line in listing is not in window, so we have to set
        // a new one.  We can just scroll up one line.
        //
        dl->setLast(dl->lastLine()->next());
        dl->deleteLine();

        if (CS || DL)
        {
            if (CS)
                delete_listing_line(y);
            else
            {
                clear_modeline();
                delete_screen_line(y);
                update_modeline();
            }
            move_cursor(rows()-3, 0);
            display_string(dl->lastLine()->line(), dl->lastLine()->length());
        }
        else
        {
            clear_to_end_of_screen(y);
            move_cursor(y, 0);
            DirLine *ln = dl->currLine();
            for (int i = y; i < rows()-2; i++, ln = ln->next())
                display_string(ln->line(), ln->length());
            update_modeline();
        }

        dl->saveYXPos(y, goal_column(dl));
    }
    else
    {
        //
        // Last line of listing is visible in window,
        // so firstLIne and lastLine will not change.
        //
        if (dl->atWindowTop() && dl->atWindowBot())
        {
            //
            // The last line in the window is also the first line.
            // Since we do not allow deletion of . or .., there
            // must be more viewable lines.  Scroll down half
            // a window to put more into view.
            //
            scroll_down_half_window(dl, 0);
            dl->deleteLine();
            DirLine *ln = dl->firstLine();
            for (int pos = 0; ln != dl->tail(); pos++, ln = ln->next())
                ;
            dl->saveYXPos(pos, goal_column(dl));
            move_cursor(pos+1, 0);
            clear_to_end_of_line();
            move_cursor(pos, dl->savedXPos());
        }
        else if (dl->atWindowBot())
        {
            //
            // We want to delete the last line in the window.
            //
            dl->deleteLine();
            move_cursor(y, 0);
            clear_to_end_of_line();
            dl->saveYXPos(y-1, goal_column(dl));
            move_cursor(y-1, dl->savedXPos());
        }
        else
        {
            //
            // We are in the middle of the listing.
            //
            dl->deleteLine();
            if (CS || DL)
            {
                if (CS)
                    delete_listing_line(y);
                else
                {
                    clear_modeline();
                    delete_screen_line(y);
                    update_modeline();
                }
            }
            else
            {
                clear_to_end_of_screen(y);
                move_cursor(y, 0);
                for (DirLine *ln = dl->currLine(); ln; ln = ln->next())
                    display_string(ln->line(), ln->length());
                update_modeline();
            }
            dl->saveYXPos(y, goal_column(dl));
         }
    }
    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(dl->savedYPos(), dl->savedXPos());
}
