//
// command2.C - the main command loop and some of the commands themselves.
//              Command1.C and command2.C are concatenated during the make
//              into commands.C, which consists of the main command loop and
//              all commands called from within that loop.
//
// $Id: command2.C,v 1.12 1993/09/19 20:28:52 lijewski Exp $
//
// Copyright (c) 1991, 1992, 1993 by Mike Lijewski.
//


//
// Delete the current file which is a regular file.
// Update the window appropriately.
//

static void delete_file (DirList *dl, const char *file)
{
    if (unlink(file) != -1)
        remove_listing_line(dl, dl->savedYPos());
    else
    {
        message("`deletion' failed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
}


//
// Delete the current file which is a directory.
// Update the window appropriately.
//

static void delete_directory (DirList *dl, const char *dirname)
{
    if (rmdir(dirname) != -1)
        remove_listing_line(dl, dl->savedYPos());
    else
    {
        message("`deletion' failed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
}


//
// Attempt to delete current file.
//

void delete_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    
    if (strcmp(file, ".")  == 0 || strcmp(file, "..") == 0)
    {
        message("`deletion' of `.' and `..' is not allowed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    String msg("delete `");
    msg += file;
    msg += "'? (y|n) ";

    if (is_directory(file) && yes_or_no(msg))
        delete_directory(dl, file);
    else if (yes_or_no(msg))
        delete_file(dl, file);
    else
        //
        // Do NOT do the delete.
        //
        move_cursor(dl->savedYPos(), dl->savedXPos());
    
    synch_display();
}


//
// Updates the current line in the listing and its
// representation on the screen.
//

static void update_current_line (DirList *dl, char *new_line)
{
    if (strlen(new_line) <= columns() && dl->currLine()->length() <= columns())
    {
        //
        // The most common case.
        //
        update_screen_line(dl->currLine()->line(), new_line, dl->savedYPos());
        dl->currLine()->update(&new_line);
        dl->saveYXPos(dl->savedYPos(), goal_column(dl));
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    else
    {
        //
        // Either the old or the new line must be shifted.
        // Probably not worth trying to use update_screen_line.
        //
        dl->currLine()->update(&new_line);
        dl->saveYXPos(dl->savedYPos(), goal_column(dl));
        
        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
        {
            move_cursor(dl->savedYPos(), 0);
            clear_to_end_of_line();
            display_string(new_line);
            move_cursor(dl->savedYPos(), dl->savedXPos());
        }
    }
}


//
// Rename src to dest and update window appropriately.
//

static void rename_file (DirList *dl, const char *src, const char *dest)
{
    if (rename(src, dest) != -1)
    {
        message("`rename' was successful");
        if (in_same_directory(dl, dest))
        {
            //
            // The important point to note about the following code
            // is that if dest is a directory, we need to add a "-d "
            // to the ls command as in
            //
            //   ls -ld source
            //
            String command = ls_cmd[the_sort_order()];
            if (is_directory(dest))
                command += "-d ";
            command += "'";
            command += dest;
            command += "'";

            FILE *fp = popen(command, "r");
            if (fp == 0)
                error("File %s, line %d: popen(%s) failed",
                      __FILE__, __LINE__, (const char *)command);

            char *new_line = fgetline(fp);
            if (fp == 0)
                error("File %s, line %d: fgetline() failed",
                      __FILE__, __LINE__);
            (void) pclose(fp);

            update_current_line(dl, new_line);
        }
        else
            remove_listing_line(dl, dl->savedYPos());
    }
    else
    {
        message("`rename' failed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
}


//
// Attempt to rename the current file.
//

void rename_current_file (DirList *dl)
{
    const char *from_file = get_file_name(dl);
    
    if (strcmp(from_file, ".")  == 0 || strcmp(from_file, "..") == 0)
    {
        message("`renaming' of `.' and `..' is not allowed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    const char *to_file = prompt("Rename to: ");
    
    if (to_file == 0)
    {
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*to_file == 0)
    {
        message("not a valid file name");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }

    to_file = expand_tilde(to_file);

    rename_file(dl, from_file, to_file);

    synch_display();
}


//
// Compress the current file and update the window.
//

static void compress_file (DirList *dl, const char *file)
{
    String cmd("compress");
    cmd += " -f '";
    cmd += file;
    cmd += "' 1>/dev/null 2>&1";

    message("Compressing ... ");

    if (!system(cmd))
    {
        String command(ls_cmd[the_sort_order()]);
        command += "'";
        command += file;
        command += ".Z' 2>/dev/null";

        FILE *fp = popen(command, "r");
        if (fp == 0)
            error("File %s, line %d: popen(%s) failed",
                  __FILE__, __LINE__, (const char *)command);

        char *new_line = fgetline(fp);
        if (new_line)
            //
            // Only update the line if we were able to read the line.
            // Some versions of compress do not properly return a failure
            // code if they do not compress the file.  Hence we probably
            // still have the uncompressed file lying around.
            //
            update_current_line(dl, new_line);
        (void) pclose(fp);

        message("Compressing ... done");
    }
    else
        message("`compress' failed");
}


//
// Attempt to compress current file.
//

void compress_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);

    //
    // Disallow compressing of symbollically linked files.
    //
    if (BMstrstr(&(dl->currLine()->line())[dl->savedXPos()], " -> "))
    {
        message("compress'ing symbollically linked files not allowed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    if (file[strlen(file)-1] == 'Z' && file[strlen(file) - 2] == '.')
    {
        message("file appears to already be compressed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    if (is_regular_file(file))
        compress_file(dl, file);
    else
        message("can only `compress' regular files");
    move_cursor(dl->savedYPos(), dl->savedXPos());

    synch_display();
}


//
// Compress the current file using zip and update the window.
//

static void zip_file (DirList *dl, const char *file)
{
    String cmd = "gzip";
    cmd += " -9 '";
    cmd += file;
    cmd += "' 1>/dev/null 2>&1";

    message("gzip'ing ... ");

    if (!system(cmd))
    {
        String command = ls_cmd[the_sort_order()];
        command += "'";
        command += file;
        command += ".gz' 2>/dev/null";

        FILE *fp = popen(command, "r");
        if (fp == 0)
            error("File %s, line %d: popen(%s) failed",
                  __FILE__, __LINE__, (const char *)command);

        char *new_line = fgetline(fp);
        if (new_line)
            //
            // Only update the line if we were able to read the line.
            //
            update_current_line(dl, new_line);
        (void) pclose(fp);

        message("gzip'ing ... done");
    }
    else
        message("`gzip' failed");
}


//
// Attempt to zip current file.
//

void zip_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);

    //
    // Disallow zipping of symbollically linked files.
    //
    if (BMstrstr(&(dl->currLine()->line())[dl->savedXPos()], " -> "))
    {
        message("gzip'ing symbollically linked files not allowed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    if (file[strlen(file) - 1] == 'z' &&
        file[strlen(file) - 2] == 'g' &&
        file[strlen(file) - 3] == '.')
    {
        message("file appears to already be gzip'ed");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    if (is_regular_file(file))
        zip_file(dl, file);
    else
        message("can only `gzip' regular files");

    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Change the group of the current file & update the window appropriately.
//

void chgrp_file (DirList *dl, const char *file, const char *group)
{
    const char *chgrp  = "chgrp";
    const char *args[4];

    args[0] = chgrp;
    args[1] = group;
    args[2] = file;
    args[3] = 0;

    if (execute(chgrp, args))
    {
        String command = ls_cmd[the_sort_order()];
        if (is_directory(file)) command += "-d ";
        command += "'";
        command += file;
        command += "'";

        FILE *fp = popen(command, "r");
        if (fp == 0)
            error("File %s, line %d: popen(%s) failed",
                  __FILE__, __LINE__, (const char *)command);

        char *new_line = fgetline(fp);
        if (fp == 0)
            error("File %s, line %d: fgetline() failed", __FILE__, __LINE__);
        (void) pclose(fp);

        update_current_line(dl, new_line);
    }
    else
        message("`chgrp' failed");
}


//
// Attempt to chgrp current file.
//

static void chgrp_current_file (DirList *dl)
{
    const char *file  = get_file_name(dl);
    const char *group = prompt("Change to group: ");
    
    if (group == 0)
    {
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*group == 0)
    {
        message("not a valid group");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    chgrp_file(dl, file, group);

    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Change the mode of the current file & update the window appropriately.
//

static void chmod_file (DirList *dl, const char *file, const char *mode)
{
    const char *chmod = "chmod";
    const char *args[4];

    args[0] = chmod;
    args[1] = mode;
    args[2] = file;
    args[3] = 0;

    if (execute(chmod, args))
    {
        String command = ls_cmd[the_sort_order()];
        if (is_directory(file))
            command += "-d ";
        command += "'";
        command += file;
        command += "'";

        FILE *fp = popen(command, "r");
        if (fp == 0)
            error("File %s, line %d: popen(%s) failed",
                  __FILE__, __LINE__, (const char *)command);

        char *new_line = fgetline(fp);
        if (fp == 0)
            error("File %s, line %d: fgetline() failed", __FILE__, __LINE__);
        (void) pclose(fp);

        update_current_line(dl, new_line);
    }
    else
        message("`chmod' failed");
}


//
// Attempt to chmod the current file.
//

void chmod_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    const char *mode = prompt("Change to mode: ");
    
    if (mode == 0)
    {
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*mode == 0)
    {
        message("not a valid file mode");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    chmod_file(dl, file, mode);
    
    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Attempt to print current file.
//

void print_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    char *printer    = getenv("DIREDPRT");
    if (printer == 0)
        printer = "lpr";

    String cmd(printer);
    cmd += " '";
    cmd += file;
    cmd += "'";
    
    if (is_regular_file(file))
        if (!system(cmd))
            message("`%s' printed successfully", file);
        else
            message("print of `%s' failed", file);
    else
        message("`%s' isn't a regular file", file);
    
    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Uncompress the file at line y in the DirList & update
// the window appropriately.
//

static void uncompress_file (DirList *dl, const char *file)
{
    String cmd("uncompress");
    cmd += " '";
    cmd += file;
    cmd += "' 1>/dev/null 2>&1";

    message("Uncompressing ... ");

    if (!system(cmd))
    {
        char *dot = strrchr(file, '.');
        //
        // Remove .Z suffix.
        //
        *dot = 0;

        String command(ls_cmd[the_sort_order()]);
        command += "'";
        command += file;
        command += "' 2>/dev/null";

        FILE *fp = popen(command, "r");
        if (fp == 0)
           error("File %s, line %d: popen(%s) failed",
                  __FILE__, __LINE__, (const char *)command);            

        char *new_line = fgetline(fp);
        if (new_line)
            //
            // Only update the line if we were able to read the line.
            // Some versions of compress do not properly return a failure
            // code if they do not uncompress the file.  Hence we probably
            // still have the .Z file lying around.
            //
            update_current_line(dl, new_line);
        (void) pclose(fp);

        message("Uncompressing ... done");
    }
    else
        message("`uncompress' failed");
}


//
// Attempt to uncompress current file.
//

void uncompress_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    
    if (file[strlen(file)-1] != 'Z' && file[strlen(file)-1] != '.')
    {
        message("can only `uncompress' compressed files");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    uncompress_file(dl, file);

    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Search forward for file or directory matching string. The search
// always starts with the file following the current file.  If a
// match occurs, the cursor is placed on the new current file, with
// that line placed at the top of the window.
//

void search_forward (DirList *dl)
{
    const char *str = prompt("Search forward for: ");
    if (str == 0)
    {
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }                  
    
    DirLine *ln = dl->currLine();
    if (ln == dl->tail())
    {
        message("no match");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        return;
    }
    
    int pos = dl->savedYPos()+1;
    DirLine *found = search(ln->next(), str, &pos);
    if (found)
    {
        if (dl->currLine()->length() > columns())
            rightshift_current_line(dl);
        dl->setCurrLine(found);
    }
    
    if (found && pos < rows()-2)
    {
        //
        // We found a match in our current window.
        // Only need to update the cursor position.
        //
        dl->saveYXPos(pos, goal_column(dl));
        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(pos, dl->savedXPos());
    }
    else if (found)
    {
        //
        // We found a match, but it is out of our current view.
        // Place the matched line at the top of the window.
        //
        ln = found;
        dl->setFirst(ln);

        clear_display_area();

        for (int i = 0; i < rows() - 2 && ln; i++, ln = ln->next())
            display_string(ln->line(), ln->length());

        if (ln)
            dl->setLast(ln->prev());
        else
            dl->setLast(dl->tail());

        dl->saveYXPos(0, goal_column(dl));

        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(0, dl->savedXPos());
    }
    else
    {
        message("no match");
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    
    synch_display();
}


//
// Search backward for file or directory matching string. The search
// always starts with the file following the current file.  If a
// match occurs, the cursor is placed on the new current file, with
// that line placed at the top of the window.
//

void search_backward (DirList *dl)
{
    const char *str = prompt("Search backward for: ");
    if (str == 0)
    {
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    DirLine *ln = dl->currLine();
    if (ln == dl->head())
        return;
    
    int pos = dl->savedYPos()-1;
    DirLine *found = search(ln->prev(), str, &pos, 0);
    if (found)
    {
        if (dl->currLine()->length() > columns())
            rightshift_current_line(dl);
        dl->setCurrLine(found);
    }
    
    if (found && pos >= 0)
    {
        //
        // We found a match in our current window.
        // Only need to update the cursor position.
        //
        dl->saveYXPos(pos, goal_column(dl));
        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(pos, dl->savedXPos());
    }
    else if (found)
    {
        //
        // We found a match, but it is out of our
        // current view.  Place the matched line at the top
        // of the window.  Since we are searching backwards
        // and the match is out of our present window, we
        // must have a whole new window to display.
        //
        dl->setFirst(ln = found);

        clear_display_area();
        
        for (int i = 0; i < rows() - 2 && ln->next(); i++, ln = ln->next())
            display_string(ln->line(), ln->length());
        
        dl->setLast(ln->prev());
        dl->saveYXPos(0, goal_column(dl));

        if (dl->currLine()->length() > columns())
            leftshift_current_line(dl);
        else
            move_cursor(0, dl->savedXPos());
    }
    else
    {
        message("no match");
        move_cursor(dl->savedYPos(), dl->savedXPos());
    }
    synch_display();
}


//
// The help file is maintained as an array of Strings.  It
// must be initialized after the KeyMap is calculated and sorted
// and before help() is called.
//
static String* HelpFile;


//
// Dimension of HelpFile
//
static int HelpFileDim;


//
// Initializes the help file.
//

void initialize_helpfile (const KeyMap& keymap)
{
    const int DIM = keymap.entries();

    HelpFile    = ::new String[DIM];
    HelpFileDim = DIM;

    int longest_prefix = 0;

    for (int i = 0; i < DIM; i++)
    {
        HelpFile[i] = decode(keymap[i]->sequence);
        if (HelpFile[i].length() > longest_prefix)
            longest_prefix = HelpFile[i].length();
    }

    //
    // Add two more spaces for easy reading.
    //
    longest_prefix += 2;

    for (i = 0; i < DIM; i++)
    {
        for (int pad = longest_prefix - HelpFile[i].length(); pad; pad--)
            HelpFile[i] += " ";
        HelpFile[i] += keymap[i]->help_msg;
    }
}
    

//
// Help message for the message window when displaying help.
//
const char *const HelpMessage[] =
{
    "Space scrolls forward.  Other keys quit.",
    "Space forward, Backspace/Delete backward.  Other keys quit.",
    "Backspace/Delete scrolls backward.  Other keys quit."
};


//
// Give some help.  Deal with SIGWINCH and SIGTSTP.
//

void help (DirList *dl)
{
    update_modeline("----- HELP");

    int position = 0;
    char key;
    do
    {
        clear_display_area();
        
        for (int i = 0; i < rows() - 2 && i + position < HelpFileDim; i++)
            display_string(HelpFile[position+i],HelpFile[position+i].length());

        clear_message_line();

        if (position + rows() -2 >= HelpFileDim)
            //
            // The tail of the help message.
            //
            (void) fputs(HelpMessage[2], stdout);
        else if (position == 0)
            //
            // The head of the help message.
            //
            (void) fputs(HelpMessage[0], stdout);
        else
            //
            // Somewhere in between.
            //
            (void) fputs(HelpMessage[1], stdout);

        synch_display();

        if (read(0, &key, 1) < 0 // Assume fails only when errno == EINTR.
#ifdef SIGWINCH
            || win_size_changed
#endif
            )
        {
#ifdef SIGWINCH
            if (win_size_changed)
            {
                win_size_changed = 0;
                adjust_window();
                redisplay();
            }
#endif
        }
        else if (key == ' ')
        {
            if (position >= HelpFileDim - 1)
                goto finished;
            position += rows() - 2;
        }
        else if (key == *BC || key == *DC)
        {
            if (position == 0)
                goto finished;
            position -= rows() - 2;
        }
        else 
            goto finished;  // Return to the listing.
    }
    while (position < HelpFileDim - 1);

  finished:

    update_modeline(modeline_prefix, dl->name());
    redisplay();
}


//
// Reread the current directory and put it up fresh on the screen.
// We attempt to put the cursor back onto the previous current file,
// if that file still exists.
//

void reread_current_directory (DirList * &dlr)
{
    char *dirname = new char[strlen(dlr->name()) + 1];
    (void) strcpy(dirname, dlr->name());
    const char *old_current_file = get_file_name(dlr);

    //
    // Position in old DirList of old_current_file.
    //
    int old_pos;
    DirLine *ln = dlr->head();
    for (old_pos = 0; ln != dlr->currLine(); ln = ln->next(), old_pos++)
        ;

    delete dir_stack->pop();

    DirList *dl = get_directory_listing(dirname);
    if (dl == 0)
        error("File %s, line %d: couldn't read directory `%s'",
              __FILE__, __LINE__, dirname);
    //
    // Update dir_list in read_commands.
    //
    dlr = dl;

    dir_stack->push(dl);

    ln = dl->head();
    int pos = 0;
    const char *name;

    while (ln)
    {
#ifndef NO_SYMLINKS
        if ((name = BMstrstr(ln->line(), " -> ")) != 0)
            //
            // We have a symbolic link.
            //
            --name;
        else
#endif
            name = ln->line() + ln->length();

        for (int namelen = 0; isspace(*name) == 0; name--, namelen++)
            ;

        if (strncmp(name + 1, old_current_file, namelen) == 0)
            break;

        ln = ln->next();
        pos++;
    }

    if (ln == 0)
    {
        //
        // `old_current_file' doesn't exist anymore.  We must
        // choose another file to make current.  This can only
        // happen when the file was removed by some means unknown
        // to dired -- the infamous `!' command.
        //
        ln = dl->head();
        for (pos = 0; pos < old_pos && ln; ln = ln->next(), pos++)
            ;
        if (ln == 0)
        {
            //
            // This can happen if the user has does something dastardly such
            // as a `!' command which removes part of the tail of the listing
            // that contained the current file.
            //
            ln = dl->tail();
            --pos;
        }
    }

    dl->setCurrLine(ln);

    if (pos < rows() - 2)
    {
        //
        // Display starting at the head.
        //
        ln = dl->head();
        dl->setFirst(ln);
        dl->saveYXPos(pos, goal_column(dl));
    }
    else
    {
        //
        // Center dl->currLine.
        //
        ln = dl->currLine();
        for (int i = 0; i < (rows() - 1)/2; i++, ln = ln->prev())
            ;
        dl->setFirst(ln);
        dl->saveYXPos((rows() - 1)/2, goal_column(dl));
    }

    clear_display_area();
    
    for (int i = 0; i < rows() - 2 && ln; ln = ln->next(), i++)
        display_string(ln->line(), ln->length());

    if (ln)
        dl->setLast(ln->prev());
    else
        dl->setLast(dl->tail());
    
    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(dl->savedYPos(), dl->savedXPos());
    
    synch_display();
}


//
// Unzip the file at line y in the DirList & update the window appropriately.
//

void unzip_file (DirList *dl, const char *file)
{  
    String cmd = "gunzip";
    cmd += " '";
    cmd += file;
    cmd += "' 1>/dev/null 2>&1";

    message("gunzip'ing ... ");

    if (!system(cmd))
    {
        char *dot = strrchr(file, '.');
        //
        // Remove .z suffix.
        //
        *dot = 0;

        String command = ls_cmd[the_sort_order()];
        command += "'";
        command += file;
        command += "' 2>/dev/null";

        FILE *fp = popen(command, "r");
        if (fp == 0)
           error("File %s, line %d: popen(%s) failed",
                  __FILE__, __LINE__, (const char *)command);            

        char *new_line = fgetline(fp);
        if (new_line)
            //
            // Only update the line if we were able to read the line.
            //
            update_current_line(dl, new_line);
        (void) pclose(fp);

        message("gnzip'ing ... done");
    }
    else
        message("`gunzip' failed");
}


//
// Attempt to unzip current file.
//

static void unzip_current_file (DirList *dl)
{
    const char *file = get_file_name(dl);
    int len = strlen(file);
    
    if (len < 4 || strcmp(&file[len-3], ".gz"))
    {
        message("can only `unzip' gzip'ed files");
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    
    unzip_file(dl, file);

    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Expand any percent signs in cmd to file.  The result is stored
// in volatile storage.  We must be careful about %s embedded in file itself.
//

static const char *expand_percent (const char *cmd, const char *file)
{
    static String expanded_cmd;
    const char *pos = strchr(cmd, '%');

    if (!pos)
        return cmd;

    expanded_cmd = "";
    for (const char *oldpos = cmd; pos; oldpos=pos+1, pos=strchr(oldpos, '%'))
    {
        expanded_cmd += String(oldpos, pos - oldpos);
        expanded_cmd += file;
    }
    expanded_cmd += oldpos;

    return (const char *)expanded_cmd;
}


//
// Execute a shell command.  % is expanded to the current filename.
// If *cmd == 0, start up a shell.
// If *cmd == !, reexecute most recent shell command.
//

void shell_command (DirList *dl)
{
    static String original_cmd;
    static String saved_cmd;
    static String saved_shell;
    const char *file = get_file_name(dl);
    const char *cmd  = prompt("!");

    if (cmd == 0)
    {
        //
        // Command aborted.
        //
        move_cursor(dl->savedYPos(), dl->savedXPos());
        synch_display();
        return;
    }
    else if (*cmd == 0)
    {
        //
        // Start up a shell.
        //
        if (saved_shell == "")
            saved_shell = getenv("SHELL");
        if (saved_shell == "")
            saved_shell = "sh";
        saved_cmd = original_cmd = saved_shell;

        const char *slash = strrchr(saved_shell, '/');
        const char *args[2];
        args[0] = slash ? slash + 1 : (const char *)saved_shell;
        args[1] = 0;

        message("Starting interactive shell ...");

        cursor_wrap();
        synch_display();
        unsetraw();
        unset_signals();

        execute(saved_shell, args, 0);

        set_signals();
        setraw();
    }
    else if (*cmd == '[')
    {
        //
        // Re-expand previous original command, if it contains a %.
        // Otherwise, re-execute previous saved command since the original
        // and saved will be the same.
        //
        if (original_cmd != "")
        {
            //
            // Expand the %.
            //
            saved_cmd = expand_percent(original_cmd, file);
            message(saved_cmd);
            cursor_wrap();

            exec_with_system(saved_cmd);
        }
        else
        {
            message("No Previous Shell Command");
            move_cursor(dl->savedYPos(), dl->savedXPos());
            synch_display();
            return;
        }
    }
    else if (*cmd == '!')
    {
        //
        //  Re-execute previously saved command.
        //
        if (saved_cmd != "")
        {
            message(saved_cmd);
            cursor_wrap();

            exec_with_system(saved_cmd);
        }
        else
        {
            message("No Previous Shell Command");
            move_cursor(dl->savedYPos(), dl->savedXPos());
            synch_display();
            return;
        }
    }
    else
    {
        //
        // Expand and then execute command.
        //
        original_cmd = cmd;
        saved_cmd    = expand_percent(original_cmd, file);
        message(saved_cmd);
        cursor_wrap();

        exec_with_system(saved_cmd);
    }
    redisplay();
}


//
// Query the user for a new sorting order - expects to read a single
// character - and change the sort order to that value, if it is a
// valid value.  Deals with SIGWINCH and SIGTSTP.
//

void change_sort_order (DirList *dl)
{
    const char *msg = "Set sort order to? [a, c, t, u]: ";
    message(msg);

    char key;
    while (1)
        if (read(0, &key, 1) < 0 // assume fails only when errno == EINTR
#ifdef SIGWINCH
            || win_size_changed
#endif
            )
        {
#ifdef SIGWINCH
            if (win_size_changed)
            {
                win_size_changed = 0;
                adjust_window();
                redisplay();
            }
#endif
            //
            // Redisplay the message - termstop takes care of the rest
            // in the case of SIGTSTP.
            //
            message(msg);
        }
    else
        switch(key)
        {
          case 'a':
            set_sort_order(ALPHABETICALLY);
            message("Sorting `alphabetically' now.");
            goto end;
          case 'c':
            set_sort_order(INODE_CHANGE_TIME);
            message("Sorting by `inode-change time' now.");
            goto end;
          case 't':
            set_sort_order(MODIFICATION_TIME);
            message("Sorting by `modification time' now.");
            goto end;
          case 'u':
            set_sort_order(ACCESS_TIME);
            message("Sorting by `access time' now.");
            goto end;
          default:
            message("Sort order not changed.");
            goto end;
        }
  end:
    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// Pop a directory.  If this is the last directory in the directory
// stack we return 1 which signifies that we must exit the command
// execution loop.
//

int pop_a_directory (DirList*& dl)
{
    delete dir_stack->pop();
    
    if (dir_stack->isEmpty())
        return 1;

    dl = dir_stack->top();
    
    //
    // Update our current directory.
    //
    if (chdir(dl->name()) < 0)
        error("File %s, line %d: couldn't chdir() to `%s'",
              __FILE__, __LINE__, dl->name());
    //
    // We track the CWD and PWD variables if defined,
    // so that applications, such as emacs, which use them
    // will work properly.
    //
    if (getenv("CWD"))
    {
        static String str;
        static String ostr;
        str =  "CWD=";
        str += dl->name();
        if (putenv(str) < 0)
            error("File %s, line %d: putenv(%s) failed.",
                  __FILE__, __LINE__, dl->name());
        ostr = str;
    }
    if (getenv("PWD"))
    {
        static String str;
        static String ostr;
        str =  "PWD=";
        str += dl->name();
        if (putenv(str) < 0)
            error("File %s, line %d: putenv(%s) failed.",
                  __FILE__, __LINE__, dl->name());
        ostr = str;
    }
    
    update_modeline(modeline_prefix, dl->name());
    redisplay();

    return 0;
}


//
// Bye!
//

void quit (DirList*)
{
    deinit_screen_and_kbdr();
    exit(0);
}


//
// Print out the version string.
//

void display_version_string (DirList* dl)
{
    message(version);
    move_cursor(dl->savedYPos(), dl->savedXPos());
    synch_display();
}


//
// The command loop.
//

void read_commands (DirList *dir_list, const KeyMap& keymap)
{
    CommandCode cmd;

    for (;;)
    {
        if (dir_stack->isEmpty())
            return;
        
        cmd = get_command(keymap);

        if (message_window_dirty)
        {
            clear_message_line();
            move_cursor(dir_list->savedYPos(), dir_list->savedXPos());
            synch_display();
            message_window_dirty = 0;
        }

        switch (cmd)
        {
          case CURSOR_DOWN: cursor_down_one_line(dir_list); break;

          case CURSOR_UP: cursor_up_one_line(dir_list); break;

          case PAGE_UP: scroll_up_full_window(dir_list); break;

          case PAGE_DOWN: scroll_down_full_window(dir_list); break;

          case HALFPAGE_UP: scroll_up_half_window(dir_list); break;

          case HALFPAGE_DOWN: scroll_down_half_window(dir_list); break;

          case TOP_OF_FILE: goto_first(dir_list); break;

          case BOTTOM_OF_FILE: goto_last(dir_list); break;

          case EDIT_FILE: edit_current_file(dir_list, keymap); break;

          case PAGE_FILE: page_current_file(dir_list, keymap); break;

          case COPY_FILE: copy_current_file(dir_list); break;

          case DELETE_FILE: delete_current_file(dir_list); break;

          case RENAME_FILE: rename_current_file(dir_list); break;

          case COMPRESS_FILE: compress_current_file(dir_list); break;

          case EDIT_DIRECTORY:

              edit_prompted_for_directory(dir_list, keymap); break;

          case CHGRP_FILE: chgrp_current_file(dir_list); break;

          case HELP: help(dir_list); break;

          case LINK_FILE: link_current_file(dir_list); break;

          case CHMOD_FILE: chmod_current_file(dir_list); break;

          case CHANGE_SORT_ORDER: change_sort_order(dir_list); break;

          case PRINT_FILE: print_current_file(dir_list); break;

          case REREAD_DIRECTORY: reread_current_directory(dir_list); break;

#ifndef NO_SYMLINKS
          case SYMLINK_FILE: symlink_current_file(dir_list); break;
#endif

          case GUNZIP_FILE: unzip_current_file(dir_list); break;

          case UNCOMPRESS_FILE: uncompress_current_file(dir_list); break;

          case SEARCH_FORWARD: search_forward(dir_list); break;

          case SEARCH_BACKWARD: search_backward(dir_list); break;

          case SHELL_COMMAND: shell_command(dir_list); break;

          case GZIP_FILE: zip_current_file(dir_list); break;

          case REDISPLAY: redisplay(); break;

          case VERSION: display_version_string(dir_list); break;

          case POP_THIS_DIRECTORY:

              if (pop_a_directory(dir_list))
                  return;
              break;

          case QUIT: quit(dir_list); break;

          case UNKNOWN: ding(); break;

          default:
              error("file %s, line %d, how did this happen?",
                    __FILE__, __LINE__);
              break;
              
        }
    }
}
