//
// dired - a directory editor modelled after GNU Emacs Dired mode.
//
// Written in C++ using the termcap library
//
// $Id: dired.C,v 1.13 1993/08/24 19:19:35 lijewski Exp $
//
// Copyright (c) 1991, 1992, 1993 by Mike Lijewski.
//

#include <ctype.h>
#include <new.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>

#ifdef apollo
extern "C" int chdir (const char *);
#else
#include <unistd.h>
#endif

#include "String.h"
#include "dired.h"
#include "display.h"

//
// This function isn't defined under ObjectCenter 2.0 on my Sun.
// ... nor is it defined with Apollo C++.
//
#if (defined(sun) && defined(__CLCC__)) || defined(apollo)
extern "C" int siginterrupt (int, int);
#endif


//
// the stack of directories being edited
//
DirStack *dir_stack;


//
// The default sort order is ALPHABETICALLY.
//
sort_order how_to_sort = ALPHABETICALLY;


//
// commands to get long directory listings:
//
//    ls_cmd[0] gives alphabetical listing
//    ls_cmd[1] gives listing sorted by modification time
//    ls_cmd[2] gives listing sorted by access time
//    ls_cmd[3] gives listing sorted by inode-change time
//
const char *const ls_cmd[4] =
#ifdef L_AND_G
  { "ls -agl ", "ls -aglt ", "ls -agltu ", "ls -acglt " };
#else
  { "ls -al ", "ls -alt ", "ls -altu ", "ls -aclt " };
#endif


//
// the modeline prefix
//
const char *const modeline_prefix = "----- Dired: ";


//
// has window size changed -- should really be a sig_atomic_t
//
int win_size_changed;


//
// exception handler for new - called once in main
//

static void free_store_exception (void)
{
    error("exiting, memory exhausted, sorry");
}


//
// Edit the given directory. We must have read and execute permission
// to edit a directory; calling routines must guarantee this.
//

void dired (const char *dirname, const KeyMap& keymap)
{
    if (chdir(dirname) < 0)
        error("File %s, line %d: couldn't chdir() to `%s'",
              __FILE__, __LINE__, dirname);

    char *fullname = current_directory();

    if (fullname == 0)
        error("File %s, line %d: current_directory() failed on `%s'.",
              __FILE__, __LINE__, dirname);

    DirList *dir_list = get_directory_listing(fullname);

    if (dir_list == 0)
        error("File %s, line %d: couldn't read directory `%s'",
              __FILE__, __LINE__, fullname);
    //
    // We track the CWD and PWD variables, if they are defined, so that
    // applications such as emacs which use them will work properly.
    //
    if (getenv("CWD"))
    {
        static String str;
        static String ostr;
        str =  "CWD=";
        str += fullname;
        if (putenv(str) < 0)
            error("File %s, line %d: putenv(%s) failed.",
                  __FILE__, __LINE__, fullname);
        ostr = str;
    }

    if (getenv("PWD"))
    {
        static String str;
        static String ostr;
        str =  "PWD=";
        str += fullname;
        if (putenv(str) < 0)
            error("File %s, line %d: putenv(%s) failed.",
                  __FILE__, __LINE__, fullname);
        ostr = str;
    }

    dir_stack->push(dir_list);
    initial_listing(dir_list);
    update_modeline(modeline_prefix, fullname);
    clear_message_line();
    dir_list->saveYXPos(0, goal_column(dir_list));

    if (dir_list->currLine()->length() > columns())    
        leftshift_current_line(dir_list);
    else
        move_cursor(dir_list->savedYPos(), dir_list->savedXPos());

    synch_display();

    read_commands(dir_list, keymap);  // main command loop
}


//
// This is the sort routine we pass to qsort for sorting
// an array of EncodedCommand*s -- the internal representation
// of a KeyMap.
//

static int ECcmp (const void *a, const void *b)
{
    const char* str1 = (*(EncodedCommand**)a)->sequence;
    const char* str2 = (*(EncodedCommand**)b)->sequence;
    return strcmp(str1, str2);
}


static void usage (void)
{
    (void)printf("dired [-t | -u | -c] [dirname]\n");
}


int main (int, char *argv[])
{
    //
    // Process options - the only options we accept are -t -u or -c.
    //
    while (*++argv && **argv == '-')
    {
        if (strcmp(*argv, "-t") == 0)
        {
            set_sort_order(MODIFICATION_TIME);
            continue;
        }
        else if (strcmp(*argv, "-u") == 0)
        {
            set_sort_order(ACCESS_TIME);
            continue;
        }
        else if (strcmp(*argv, "-c") == 0)
        {
            set_sort_order(INODE_CHANGE_TIME);
            continue;
        }
        else
        {
            (void)printf("unknown option `%s'\n", *argv);
            usage();
            exit(1);
        }
    }

    if (!isatty(0) || !isatty(1))
    {
        (void)fputs("stdin & stdout must be terminals\n", stderr);
        exit(EXIT_FAILURE);
    }

    //
    // If you do not have SIGINTERRUPT then signals almost surely interrupt
    // read.  If you do have it, you will need this to ensure that signals
    // interrupt slow system calls -- we are only interested in read.
    //
#ifdef SIGINTERRUPT
#ifdef SIGTSTP
    if (siginterrupt(SIGTSTP, 1) < 0) 
    {
        perror("siginterrupt(SIGTSTP)");
        exit(1);
    }
#endif
#ifdef SIGWINCH
    if (siginterrupt(SIGWINCH, 1) < 0)
    {
        perror("siginterrupt(SIGWINCH)");
        exit(1);
    }
#endif
#endif /*SIGINTERRUPT*/

    set_new_handler(free_store_exception);
    initialize();
    set_signals();

    KeyMap ourKeyMap;

    initialize_keymap_from_file(ourKeyMap, ".diredrc");

    //
    // Always activate up and down arrow, PgUp and PgDn,
    // and Home and End keys, if defined.
    //
    if (KU)  ourKeyMap.add(CURSOR_UP,      encode(KU));
    if (KD)  ourKeyMap.add(CURSOR_DOWN,    encode(KD));
    if (KH)  ourKeyMap.add(TOP_OF_FILE,    encode(KH));
    if (KHH) ourKeyMap.add(BOTTOM_OF_FILE, encode(KHH));
    if (KN)  ourKeyMap.add(PAGE_UP,        encode(KN));
    if (KP)  ourKeyMap.add(PAGE_DOWN,      encode(KP));

    //
    // Check the sanity of the supplied KeyMap.  We first sort the
    // KeyMap and then do a pairwise search for proper substrings.
    //
    qsort(ourKeyMap.Map(),ourKeyMap.entries(),sizeof(EncodedCommand**),ECcmp);

    int conflicts = 0;
    for (int i = 1; i < ourKeyMap.entries(); i++)
    {
        if (strstr(ourKeyMap[i]->sequence, ourKeyMap[i-1]->sequence))
        {
            conflicts++;
            String msg(decode(ourKeyMap[i-1]->sequence));
            msg += " conflicts with ";
            msg += decode(ourKeyMap[i]->sequence);
            message(msg);
            sleep(2);
        }
    }
    if (conflicts)
        error("sorry, no conflicts allowed -- please fix your .diredrc");

    initialize_helpfile(ourKeyMap);

    DirStack directories;

    dir_stack = &directories;

    char* dirname;
    if (*argv == 0)
    {
        dirname = ".";
        if (!is_directory(dirname))
            error("File %s, line %d: `%s' isn't a directory",
                  __FILE__, __LINE__, dirname);
        if (!read_and_exec_perm(dirname))
            error("File %s, line %d: need read & exec permission to edit `%s'",
                  __FILE__, __LINE__, dirname);

        dired(dirname, ourKeyMap);
    }
    else
        do
        {
            dirname = *argv++;
            if (!is_directory(dirname))
            {
                if (*argv)
                {
                    String msg("`");
                    msg += dirname;
                    msg += "' isn't a directory, skipping ...";
                    message(msg);
                    sleep(2);
                }
                else
                    error("File %s, line %d: `%s' isn't a directory",
                          __FILE__, __LINE__, dirname);
                continue;
            }
            if (!read_and_exec_perm(dirname))
            {
                if (*argv)
                {
                    String msg("need read & execute permission to edit `");
                    msg += dirname;
                    msg += "', skipping ...";
                    message(msg);
                    sleep(2);
                }
                else
                    error("File %s, line %d: need r & x permission to edit `%s'",
                          __FILE__, __LINE__, dirname);
                continue;
            }

            dired(dirname, ourKeyMap);
        }
        while (*argv);

    deinit_screen_and_kbdr();

    return 0;
}
