//
// keys.C - contains implementation of virtual keyboard for dired.
//
// $Id: keys.C,v 1.12 1993/06/20 19:01:50 lijewski Exp $
//
// Copyright (c) 1993 by Mike Lijewski.
//

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include "Queue.h"
#include "String.h"
#include "dired.h"
#include "keys.h"


static Command Commands[] = 
{
    { CURSOR_DOWN,        "cursor_down"         },
    { CURSOR_UP,          "cursor_up"           },
    { PAGE_UP,            "page_up"             },
    { PAGE_DOWN,          "page_down"           },
    { HALFPAGE_UP,        "halfpage_up"         },
    { HALFPAGE_DOWN,      "halfpage_down"       },
    { TOP_OF_FILE,        "top_of_file"         },
    { BOTTOM_OF_FILE,     "bottom_of_file"      },
    { EDIT_FILE,          "edit_file"           },
    { PAGE_FILE,          "page_file"           },
    { COPY_FILE,          "copy_file"           },
    { DELETE_FILE,        "delete_file"         },
    { RENAME_FILE,        "rename_file"         },
    { COMPRESS_FILE,      "compress_file"       },
    { EDIT_DIRECTORY,     "edit_directory"      },
    { CHGRP_FILE,         "chgrp_file"          },
    { HELP,               "help"                },
    { LINK_FILE,          "link_file"           },
    { CHMOD_FILE,         "chmod_file"          },
    { CHANGE_SORT_ORDER,  "change_sort_order"   },
    { PRINT_FILE,         "print_file"          },
    { REREAD_DIRECTORY,   "reread_directory"    },
#ifndef NO_SYMLINKS
    { SYMLINK_FILE,       "symlink_file"        },
#endif
    { GUNZIP_FILE,        "gunzip_file"         },
    { UNCOMPRESS_FILE,    "uncompress_file"     },
    { SEARCH_FORWARD,     "search_forward"      },
    { SEARCH_BACKWARD,    "search_backward"     },
    { SHELL_COMMAND,      "shell_command"       },
    { VERSION,            "version"             },
    { GZIP_FILE,          "gzip_file"           },
    { REDISPLAY,          "redisplay"           },
    { POP_THIS_DIRECTORY, "pop_this_directory"  },
    { QUIT,               "quit"                }
};


const int NCommands = sizeof(Commands) / sizeof(Commands[0]);


const char* Help[UNKNOWN] =
{
    "Forward  one line.",
    "Backward one line.",
    "Forward  one window.",
    "Backward one window.",
    "Forward  one half-window.",
    "Backward one half-window.",
    "Go to first line of listing.",
    "Go to last line of listing.",
    "Edit the current file (w/ $EDITOR) or directory.",
    "View current file with $PAGER (default `more').",
    "Copy current file - prompts for destination file.",
    "Delete current file - prompts for affirmation.",
    "Rename current file.",
    "Compress current file.",
    "Prompt for and edit a directory.",
    "Change the group of the current file.",
    "Display this help.",
    "Link current file to another file.",
    "Change the mode of the current file.",
    "Prompt for a new sorting order (a, c, t, or u).",
    "Print current file with $DIREDPRT (default `lpr').",
    "Rereads the current directory and updates the display.",
#ifndef NO_SYMLINKS
    "Create symbolic link to current file.",
#endif
    "Unzip current file using GNU unzip (gunzip).",
    "Uncompress current file.",
    "Search forward for string.",
    "Search backward for string.",
    "Runs a shell command.",
    "Print out version string.",
    "Zip current file using GNU zip (gzip).",
    "Repaint screen.",
    "Back up directory tree if possible, else quit.",
    "Exit immediately."
};


//
// Implementation of KeyMap.
//

KeyMap::KeyMap (void)
{
    map = ::new EncodedCommand*[(size = NCommands)];
    nentries = 0;
}


KeyMap::~KeyMap (void)
{
    for (int i = 0; i < nentries; i++) delete map[i];
    delete [] map;
}


KeyMap::KeyMap (const KeyMap& k)
{
    nentries = k.nentries;
    map = ::new EncodedCommand*[(size = k.size)];
    for (int i = 0; i < nentries; i++)
        map[i] = new EncodedCommand(k.map[i]->keycode, k.map[i]->sequence);
}


KeyMap& KeyMap::operator= (const KeyMap& k)
{
    if (this != &k)
    {
        for (int i = 0; i < nentries; i++) delete map[i];
        nentries = k.nentries;
        map = ::new EncodedCommand*[(size = k.size)];
        for (i = 0; i < nentries; i++)
            map[i] = ::new EncodedCommand(k.map[i]->keycode, k.map[i]->sequence);
    }
    return *this;
}


void KeyMap::add (CommandCode code, const char* seq)
{
    if (!seq) return;

    if (nentries == size)
    {
        EncodedCommand** nmap = ::new EncodedCommand*[(size += NCommands)];
        memcpy(nmap, map, entries() * sizeof(EncodedCommand*));
        delete [] map;
        map  = nmap;
    }
    map[nentries++] = ::new EncodedCommand(code, seq);
}


void KeyMap::add (CommandCode code, const String& seq)
{
    if (nentries == size)
    {
        EncodedCommand** nmap = ::new EncodedCommand*[(size += NCommands)];
        memcpy(nmap, map, entries() * sizeof(EncodedCommand*));
        delete [] map;
        map  = nmap;
    }
    map[nentries++] = ::new EncodedCommand(code, seq);
}


//
// Returns the CommandCode code for the command; UNKNOWN if it is not a valid
// command.
//

static CommandCode command_to_keycode (const char* name)
{
    CommandCode rc = UNKNOWN;
    for (int i = 0; i < NCommands; i++)
        if (!strcmp(Commands[i].name,name))
        {
            rc = Commands[i].keycode;
            break;
        }
    return rc;
}


//
// Build the default keymap.
//

static KeyMap& default_keymap (void)
{
    static KeyMap keymap;

    keymap.add(CURSOR_DOWN,        encode("j"));
    keymap.add(CURSOR_DOWN,        encode("n"));
    keymap.add(CURSOR_DOWN,        encode("\\0x20"));
    keymap.add(CURSOR_DOWN,        encode("^N"));
    keymap.add(CURSOR_DOWN,        encode("\r"));
    keymap.add(CURSOR_UP,          encode("k"));
    keymap.add(CURSOR_UP,          encode("p"));
    keymap.add(CURSOR_UP,          encode("^P"));
    keymap.add(CURSOR_UP,          encode("^Y"));
    keymap.add(PAGE_UP,            encode("^F"));
    keymap.add(PAGE_UP,            encode("^V"));
    keymap.add(PAGE_DOWN,          encode("^B"));
    keymap.add(PAGE_DOWN,          encode("^[v"));
    keymap.add(HALFPAGE_UP,        encode("^D"));
    keymap.add(HALFPAGE_DOWN,      encode("^U"));
    keymap.add(TOP_OF_FILE,        encode("<"));
    keymap.add(TOP_OF_FILE,        encode("^[<"));
    keymap.add(BOTTOM_OF_FILE,     encode(">"));
    keymap.add(BOTTOM_OF_FILE,     encode("^[>"));
    keymap.add(EDIT_FILE,          encode("e"));
    keymap.add(EDIT_FILE,          encode("f"));
    keymap.add(PAGE_FILE,          encode("m"));
    keymap.add(PAGE_FILE,          encode("v"));
    keymap.add(COPY_FILE,          encode("c"));
    keymap.add(DELETE_FILE,        encode("d"));
    keymap.add(RENAME_FILE,        encode("r"));
    keymap.add(COMPRESS_FILE,      encode("C"));
    keymap.add(EDIT_DIRECTORY,     encode("E"));
    keymap.add(CHGRP_FILE,         encode("G"));
    keymap.add(HELP,               encode("?"));
    keymap.add(HELP,               encode("H"));
    keymap.add(LINK_FILE,          encode("L"));
    keymap.add(CHMOD_FILE,         encode("M"));
    keymap.add(CHANGE_SORT_ORDER,  encode("O"));
    keymap.add(PRINT_FILE,         encode("P"));
    keymap.add(REREAD_DIRECTORY,   encode("g"));
    keymap.add(REREAD_DIRECTORY,   encode("R"));
#ifndef NO_SYMLINKS
    keymap.add(SYMLINK_FILE,       encode("S"));
#endif
    keymap.add(GUNZIP_FILE,        encode("u"));
    keymap.add(UNCOMPRESS_FILE,    encode("U"));
    keymap.add(SEARCH_FORWARD,     encode("/"));
    keymap.add(SEARCH_BACKWARD,    encode("\\"));
    keymap.add(SHELL_COMMAND,      encode("!"));
    keymap.add(VERSION,            encode("V"));
    keymap.add(GZIP_FILE,          encode("z"));
    keymap.add(REDISPLAY,          encode("^L"));
    keymap.add(POP_THIS_DIRECTORY, encode("q"));
    keymap.add(QUIT,               encode("Q"));

    return keymap;
}


void initialize_keymap_from_file (KeyMap& keymap, const char* file)
{
    //
    // We look for file and ~/file, in that order.
    // If neither exist, we use the default keymap.
    //
    FILE* fp = fopen(file, "r");
    if (fp == 0)
    {
        char* home = getenv("HOME");
        if (home)
        {
            String pathname(home);
            if (pathname[pathname.length() - 1] != '/')
                pathname += "/";
            pathname += file;
            fp = fopen(pathname, "r");
            if (fp == 0)
            {
                keymap = default_keymap();
                return;
            }
        }
        else
        {
            keymap = default_keymap();
            return;
        }
    }

    char** tokens;
    for (char* line = fgetline(fp); line; line = fgetline(fp))
    {
        if (tokenize(line, " \t", tokens) > 1 && tokens[0][0] != '#')
        {
            //
            // Layout of ".diredrc":
            //
            //    [ \t]*commandname[ \t]+encoding
            //
            CommandCode code = command_to_keycode(tokens[0]);

            if (code != UNKNOWN)
            {
                const char* encoding = encode(tokens[1]);
                if (encoding == 0)
                {
                    //
                    // Problem building encoded command.
                    //
                    delete [] line;
                    continue;
                }
                keymap.add(code, encoding);
            }
        }
        delete [] line;
    }
    (void) fclose(fp);
}


CommandCode get_command (const KeyMap& keymap)
{
#ifdef TEMPLATES
    static Queue<char> q(5);
#else
    static Queue q(5);
#endif

    char c = read_from_keybd();

    if (c <keymap[0]->sequence[0] || c >keymap[keymap.entries()-1]->sequence[0])
        return UNKNOWN;
    else
        q.insert(c);

    int submatch, match = 0, match_index;

    do
    {
        submatch = 0;

        int index, cmp, lo = 0, hi = keymap.entries() - 1;

        while (lo <= hi)
        {
            index = (lo + hi) / 2;
            if ((cmp = strncmp(keymap[index]->sequence, q.data(), q.entries())) == 0)
			{
                if (keymap[index]->sequence.length() == q.entries())
                {
                    match_index = index;
                    match = 1;
                }
                else
                {
                    submatch = 1;
                    q.insert(read_from_keybd());
                }
                break;
            }
            else if (cmp < 0)
                lo = index + 1;
            else
                hi = index - 1;
        }
    }
    while (submatch);

    CommandCode rc = match ? keymap[match_index]->keycode : UNKNOWN;

    int entries = q.entries();
    for (int i = 0; i < entries; i++)
        //
        // Remove the characters.
        //
        (void) q.get();

    return rc;        
}
    
#ifdef TEST

static const char* keycode_to_command (CommandCode key)
{
    for (int i = 0; i < NCommands; i++)
        if (Commands[i].keycode == key)
            return Commands[i].name;
}

main()
{
    KeyMap k;

    initialize_keymap_from_file(k, "junk");

    cout << k.entries() << " entries in the keymap for `junk'\r\n";

    for (int i = 0; i < k.entries(); i++)
        cout << decode(k[i]->sequence) << "\n";

    return 0;
}

#endif /*TEST*/



        
