//
// utilities.C - utility functions
//
// $Id: utilities.C,v 1.17 1993/09/19 20:28:52 lijewski Exp $
//
// Copyright (c) 1991, 1992, 1993  by Mike Lijewski.
//


#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#ifdef _IBMR2
#include <sys/access.h>
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef COMPLETION
#ifdef apollo
#include <sys/dir.h>
#include <sys/file.h>
#include <sys/stat.h>
#define	DIRENT	struct direct
#define	S_ISDIR(m)	((m&S_IFMT) == S_IFDIR)
#define	S_ISREG(m)	((m&S_IFMT) == S_IFREG)
#define	S_ISLNK(m)	((m&S_IFMT) == S_IFLNK)
extern
{
    char* getcwd (char *str,int len);
    int	  getuid(void);
    long  strtol(const char *, char **, int);
    int	  read(int, const char *, int);
    pid_t fork(void);
    int	  close(int);
}
#else
#include <dirent.h>
#define	DIRENT	struct dirent
#endif /*apollo*/
#endif /*COMPLETION*/

#ifndef apollo
#include <unistd.h>
#endif

#ifndef R_OK
#include <sys/file.h>
#endif

#include "String.h"
#include "dired.h"
#include "display.h"
#include "keys.h"


//
// Keys used when in a prompt.
//

const char KEY_CTL_L   = '\f';  // repaint screen - CTR-L
const char KEY_TAB     = '\t';  // TAB performs filename completion
const char KEY_BKSP    = '\b';  // backspace works as expected in a prompt
const char KEY_CR      = '\r';  // carriage return
const char KEY_DEL     = 127;   // ASCII DELETE works like backspace in prompt
const char KEY_ABORT   = 0x7;   // should be '\a' but for broken compilers ...
const char KEY_CTL_U   = 0x15;  // delete complete prompt
const char KEY_CTL_W   = 0x17;  // delete word


//
// Expects a string of the form "~ ...".  Returns a new string in
// volatile storage with the user's home directory in place of the ~.
// The user's home directory is always appended in the form:
// "/usr/staff/mjlx"; a slash is not added to the end of the home
// directory string.  Returns the original string if we cannot get
// the user's home directory.  The user should not attempt to delete
// the return value.
//

const char *expand_tilde (const char *str)
{
    if (*str != '~')
        return str;

    String string(str);

    static String expansion;

    //
    // Null out `expansion'.
    //
    expansion = "";

    if (string == "~" || (string[0] == '~' && string[1] == '/'))
    {
        //
        // Replace "~" by "$HOME".
        //
        static char *home = getenv("HOME");
        if (home == 0)
        {
            struct passwd *user = getpwuid(getuid());
            if (user == NULL)
                return str;
            home = user->pw_dir;
        }
        expansion += home;
        expansion += (str + 1);
    }
    else
    {
        //
        // Assume we have a string of the form "~user" or "~user/...".
        // We must try to find the home directory of the given user.
        //
        String nstr(str+1);
        char* slash = strchr((const char*)nstr, '/');
        if (slash)
            *slash = 0;
        struct passwd *user = getpwnam((const char*)nstr);
        if (user == NULL)
            return str;
        expansion += user->pw_dir;
        if (slash)
        {
            expansion += "/";
            expansion += (slash + 1);
        }
    }

    return expansion;
} 


//
// A specialized implementation of the ANSI C function strstr
// using Boyer-Moore substring searching.  Is really fast when
// the same substring is used for many searches.
//

char* BMstrstr (const char* str, const char* substr)
{
    const  int DIM = 128;
    static int table[DIM];
    static char* old_substr;
    static int gap;

    if (str == 0 || substr == 0)
        return 0;

    int string_length = strlen(str), substr_length = strlen(substr);

    if (old_substr == 0 || strcmp(substr, old_substr))
    {
        int len = substr_length;
        delete [] old_substr;
        old_substr = ::new char[substr_length + 1];
        (void)strcpy(old_substr, substr);
        for (int i = 0; i < DIM; i++)
            table[i] = substr_length;
        for (const char* p = substr; *p; p++)
            table[*p] = --len;
        char endchar = substr[substr_length - 1];
        for (gap = 1, p = substr+substr_length-2; p != substr; gap++, p--)
            if (endchar == *p)
                break;
    }

    int index = substr_length - 1;
    
    while (index < string_length)
    {
        if (table[str[index]] == 0)
        {
            if (strncmp(&str[index-substr_length+1],substr,substr_length) == 0)
                return (char*)&str[index-substr_length+1];
            index += gap;
        }
        else
            index += table[str[index]];
    }
    return 0;
}


//
// Returns a pointer to the start of a line read from fp, or the null
// pointer if we hit eof or get an error from fgets. Exits if new fails.
// Strips the newline from the line. Caller should free memory if desired.
//

static const int FGETLINE_BUFSIZE = 80; // chunksize for calls to new

char *fgetline (FILE *fp)
{
    char *buffer = new char[FGETLINE_BUFSIZE];

    char *result = fgets(buffer, FGETLINE_BUFSIZE, fp);
    if (result == 0)
    {
        //
        // Either error or at eof.
        //
        delete [] buffer;
        return 0;
    }

    if (buffer[strlen(buffer)-1] != '\n' && !feof(fp))
    {
        //
        // Longer line than buffer can hold.
        //
        char *restofline = fgetline(fp);

        if (restofline == 0)
            return 0; // eof or error

        char *longline = new char[strlen(buffer) + strlen(restofline) + 1];
        (void) strcat(strcpy(longline, buffer), restofline);

        delete [] restofline;
        delete [] buffer;

        if (longline[strlen(longline) - 1] == '\n')
            longline[strlen(longline) - 1] = 0;

        return longline;
    }
    else
    {
        if (buffer[strlen(buffer) - 1] == '\n')
            buffer[strlen(buffer) - 1] = 0;

        return buffer;
    }
}


//
// This routine tries to determine the full pathname of the current
// directory. The pointer returned is new storage.  If the beginning
// of the directory matches "/tmp_mnt" --  the default automounter
// directory -- we strip it off.
//

char *current_directory (void)
{
    const int chunksize = 50;
    int size            = chunksize;
    char *dir           = new char[size];

    while (getcwd(dir, size) == 0)
    {
        if (errno == ERANGE)
        {
            delete [] dir;
            dir = new char[size += chunksize];
            continue;
        }
        else
            //
            // Must have got an EACCES.
            //
            return 0;
    }

    const char* AutomountDir = "/tmp_mnt";
    int AutodirLen = strlen(AutomountDir);
    if (strncmp(dir, AutomountDir, AutodirLen) ==  0)
    {
        //
        // Strip automounter prefix.
        //
        char* ndir = new char[strlen(dir) + 1 - AutodirLen];
        (void) strcpy(ndir, dir + AutodirLen);
        delete [] dir;
        dir = ndir;
    }

    return dir;
}


//
// Prints a string to the given the display, guaranteeing not
// to print more than columns characters.  If the string
// exceeds the width of the window, a ! is placed in
// the final column.  Can be called with or without the
// length of the string to be printed.  In most places in
// the code we know the exact length of the strings we
// wish to print.  Note that len has a default value
// of zero defined by the declaration in "dired.h".
// We never call this when trying to write to the last
// row on the screen.  That is the dominion of message.
//

void display_string (const char *str, size_t len)
{
    size_t string_length = len == 0 ? strlen(str) : len;

    if (string_length < columns())
    {
        (void)fputs(str, stdout);
        cursor_wrap();
    }
    else if (string_length > columns())
    {
        (void)printf("%*.*s%c", columns() - 1, columns() - 1, str, '!');
        if (!AM || XN)
            cursor_wrap();
    }
    else
    {
        (void)fputs(str, stdout);
        if (!AM || XN)
            cursor_wrap();
    }
}


//
// Get a long listing of the given directory. Returns 0 if we got other
// than a "memory exhausted" error.
//

DirList *get_directory_listing (char *dirname)
{
    message("Reading directory ... ");

    String cmd(ls_cmd[the_sort_order()]);
    cmd += dirname;
    cmd += " 2>/dev/null";

    FILE *fp = popen(cmd, "r");
    if (fp == 0)
        return 0;
    
    DirList *directory = new DirList(dirname);

    //
    // discard lines of the form:
    //
    //      total 1116
    //
    char *line = fgetline(fp);
    delete [] line;
    
    while ((line = fgetline(fp)) != 0)
        directory->add(new DirLine(&line));

    message("Reading directory ... done");


    if (!(feof(fp) && !ferror(fp)))
        directory = 0;

    (void) pclose(fp);
    return directory;
}


//
// Constructs a null-terminated vector of the words in line into 
// tokens.  The vector and its elements are in volatile storage
// which we manage here.  The return value is the number of non-null
// tokens in the vector.
//

int tokenize (const char *line, const char *separators, char**& tokens)
{
    //
    // Since strtok modifies its argument, we use a copy of line.
    //
    static char *newline;     // volatile storage of vector elements
    delete [] newline;
    newline = new char[strlen(line) + 1];
    (void)strcpy(newline, line);

    const int chunksize = 5;  // chunksize to grow by
    int size   = chunksize;   // total size of vector
    int nwords = 0;           // number of words in vector
    static char **words;      // volatile storage for the word pointers
    delete [] words;
    words = new char*[chunksize];

    if ((words[nwords++] = strtok(newline, separators)) == 0)
        return 0;
    else
    {
        while (words[nwords++] = strtok(0, separators))
            if (nwords == size)
            {
                //
                // Grow words.
                //
                char **newspace = new char*[size += chunksize];
                memcpy(newspace, words, nwords * sizeof(char*));
                delete [] words;
                words = newspace;
            }
        tokens = words;
    }
    
    return nwords - 1;  // minus one since we do not count the final NULL
}


//
// Takes a string of ASCII characters, "backslash" escapes, "numeric"
// escapes and "^X" sequences, and converts it into a string of the exact
// ASCII characters themselves.  The encoded string into returned in a
// static buffer, so if you want to save the string you must make a copy
// of it.  Returns the NULL string if there was an error.
//

const char* encode (char* command)
{
    if (!command)
        return 0;

    static const char* const EscapeMap   = "abfnrtv";
    static const char* const EscapeValue = "\a\b\f\n\r\t\v";
    static const char* const ControlMap  = "@abcdefghijklmnopqrstuvwxyz[\\]^_";

    static char* buffer;  // our static storage
    static int   buflen;

    if (strlen(command) > buflen)
    {
        delete [] buffer;
        buflen = strlen(command);
        buffer = ::new char[buflen + 1];
    }

    char* bufptr = buffer;

    for (; *command != '\0'; buffer++)
    {
        switch (*command)
        {
          case '^':
            if (command[1] == 0)
                //
                // A terminating '^' by itself.
                //
                *buffer = '^';
            else if (*++command == '?')
                //
                // ASCII DEL character.
                //
                *buffer = 0x7f;
            else
            {
                char c = *command;
                if (isupper(c)) c = tolower(c);

                const char* ctrl = strchr(ControlMap, c);
                if (ctrl == 0)
                {
                    //
                    // Invalid control sequence.
                    //
                    buffer = bufptr;
                    return 0;
                }
                *buffer = ctrl - ControlMap;
            }
            ++command;
            break;
          case '\\': // Escapes.
            const char* ctrl;
            if (*++command == 0)
                //
                // A terminating '\' by itself.
                //
                *buffer = '\\';
            else if (isdigit(*command))
            {
		// Numeric escapes we allow for:
                //
                //   octal   \0nnn
                //   hex     \0xnn
                //   decimal \nnn 
                //
                long number = strtol(command, &command, 0);
                if (255 < number)
                {
                    buffer = bufptr;
                    return 0;
                }
                *buffer = (char) number;
            }
            else if ((ctrl = strchr(EscapeMap, *command)))
            {
                *buffer = EscapeValue[ctrl - EscapeMap];
                ++command;
            }
            else
                *buffer = *command++;
            break;
          default:  // A literal character.
            *buffer = *command++;
            break;
        }
    }
    *buffer = 0;

    return buffer = bufptr;
}


//
// Decodes a string of ASCII characters, making them printable.
// Returns the printable string in volatile (static) storage,
// so if you want to save the string, you must make your own copy
// of it.
//

const char* decode (const char* str)
{
    static char* buf;
    static char* bufptr = buf;
    static int buflen;

    if (2 * strlen(str) > buflen)
    {
        delete [] bufptr;
        buflen = 2 * strlen(str);
        bufptr = buf = ::new char[buflen + 1];
    }
    
    buf = bufptr;

    for (; *str; str++)
    {
        if (isprint(*str))
            *buf++ = *str;
        else if (iscntrl(*str))
        {
            *buf++ = '^';
            *buf++ = char('A' + *str - 1);
        }
    }

    *buf = 0;

    return bufptr;
}


//
// Prints error message so it can be read.  These are the error
// functions we call once we have initialized the display.
//

void error (const char *str)
{
    move_to_message_line();
    clear_to_end_of_line();
    deinit_screen_and_kbdr();
    (void) printf(str);
    putchar('\n');
    exit(EXIT_FAILURE);
}


void error (const char *fmt, int index)
{
    move_to_message_line();
    clear_to_end_of_line();
    deinit_screen_and_kbdr();
    (void) printf(fmt, index);
    putchar('\n');
    exit(EXIT_FAILURE);
}


void error (const char *fmt, const char *file, int line)
{
    move_to_message_line();
    clear_to_end_of_line();
    deinit_screen_and_kbdr();
    (void) printf(fmt, file, line);
    putchar('\n');
    exit(EXIT_FAILURE);
}


void error (const char *fmt, const char *file, int line, const char *str)
{
    move_to_message_line();
    clear_to_end_of_line();
    deinit_screen_and_kbdr();
    (void) printf(fmt, file, line, str);
    putchar('\n');
    exit(EXIT_FAILURE);
}


//
// This routine concatenates the two strings into the modeline.  The
// modeline is displayed in standout mode if possible. We never put
// more than columns characters into the modeline.  The modeline is
// the penultimate line on the terminal screen.  It does not synch
// the display.  If head == tail == 0, we just display the old modeline.
// This happens if for some reason we had to clear the screen.
//

void update_modeline (const char *head, const char *tail)
{
    static char *oldline;
    move_to_modeline();
    enter_standout_mode();

    if (head == 0 && tail == 0)
    {
        //
        // Redisplay old modeline.
        //
        (void) fputs(oldline, stdout);
        end_standout_mode();
        return;
    }

    int len = (int)strlen(head);
    char *modeline = new char[columns() + 1];
    (void) strncpy(modeline, head, columns());
    modeline[columns()] = 0;  // Ensure that it is null-terminated.

    if (len < columns())
    {
        //
        // Write exactly columns characters to modeline.
        //
        for (int i = len; i < columns() - 1 && tail && *tail; i++, tail++)
            modeline[i] = *tail;
        if (i < columns() - 1)
        {
            modeline[i++] = ' ';
            for (; i < columns(); i++) modeline[i] = '-';
        }
        else if (tail && *tail)
            //
            // The string was overly long.  Put a ! in the last space
            // on the modeline to signify truncation.
            //
            modeline[columns() - 1] = '!';
        else
            //
            // Here len == columns-1 && there is nothing else in tail.
            //
            modeline[columns() - 1] = ' ';
    }
    else if (len > columns())
        modeline[columns() - 1] = '!';

    if (oldline)
    {
        update_screen_line(oldline, modeline, rows() - 2);
        delete [] oldline;
    }
    else
        (void) fputs(modeline, stdout);

    oldline = modeline;
    end_standout_mode();
}


//
// Returns non-zero if a directory, otherwise 0. Also returns zero on error.
//

int is_directory (const char *dir)
{
    struct stat stbuf;
    if (stat(dir, &stbuf) < 0)
        return 0;
    return S_ISDIR(stbuf.st_mode);
}


//
// Returns non-zero if a regular file, otherwise 0. Also returns zero on error.
// We consider symlinks regular, if the implementation supports them.
//

int is_regular_file (const char *file)
{
    struct stat buf;
    if (stat(file, &buf) < 0)
        return 0;
    return S_ISREG(buf.st_mode)
#ifndef NO_SYMLINKS
           || S_ISLNK(buf.st_mode)
#endif
           ;
}


//
// Returns non-zero if we have read and execute permission on the
// directory, otherwise 0. Returns 0 on error.
//

int read_and_exec_perm (const char *dir)
{
    return access(dir, R_OK | X_OK) == -1 ? 0 : 1;
}


//
// Find the column position of the first character of the filename
// in the current line of the given DirList.
//
// The straight-forward way to do this is to walk the string from its
// tail to its head until we hit some whitespace.  This presumes
// that filenames do not contain whitespace.  The one special case
// to worry about is if the file is a symbolic link.  In that case we will
// have a filename field entry of the form
//
//       Xm -> /usr/lpp/include/Xm
//

int goal_column (DirList *l)
{
    DirLine *line = l->currLine();
    const char *tmp;

#ifndef NO_SYMLINKS
    if ((tmp = BMstrstr(line->line(), " -> ")) != 0)
        //
        // We have a symbolic link.
        //
        --tmp;
    else
#endif
        tmp = line->line() + line->length() - 1;

    while(!isspace(*tmp))
        --tmp;

    return tmp - line->line() + 1;
}


//
// Returns the filename of the current line of the DirList in volatile storage.
// If we have a symbolic link, we return the link not the file pointed to.
//

const char *get_file_name (DirList *dl)
{
    static String file;
    file = &(dl->currLine()->line())[goal_column(dl)];

    //
    // Do we have a symbolic link?
    //
    char *result = BMstrstr(file, " -> ");
    if (result)
        *result = 0;

    return file;
}


//
// This routine redisplays the DirList at the top of our stack.  It
// assumes that the physical screen has become corrupted, clearing
// each line before writing to it.
//

void redisplay (void)
{
    DirList *dl = dir_stack->top();

    DirLine *ln = dl->firstLine();

    clear_display_area();
    
    for (int i = 0; i < rows() - 2 && ln; i++, ln = ln->next())
        display_string(ln->line(), ln->length());

    move_to_modeline();
    clear_to_end_of_line();
    update_modeline();
    clear_message_line();

    if (dl->currLine()->length() > columns())
        leftshift_current_line(dl);
    else
        move_cursor(dl->savedYPos(), dl->savedXPos());

    synch_display();
}


//
// forward declaration
//
static void adjust_window (void);

//
// Displays the message in standout mode and waits until a character
// is typed.  Deals with SIGWINCH and SIGTSTP.
//

void eat_a_character (const char *msg)
{
    enter_standout_mode();
    message(msg);
    end_standout_mode();
    char c;
    while (1)
        if (read(0, &c, 1) < 0 // assume fails only when errno == EINTR
#ifdef SIGWINCH
            || win_size_changed
#endif
            )
        {
#ifdef SIGWINCH
            if (win_size_changed)
            {
                win_size_changed = 0;
                adjust_window();
                redisplay();
            }
#endif
            //
            // Must redisplay the message.
            //
            enter_standout_mode();
            message(msg);
            end_standout_mode();
        }
        else
            return;
}


//
// Reads a character from the keyboard.  It only returns when it has
// successfully read a character.  So if we get suspended while waiting
// to read a character, the signal handler will redisplay the screen and
// we will still be here waiting to read a character.  This is only used
// by the main command loop so that we know that redisplay will do all
// the necessary redisplay operations.
//

char read_from_keybd (void)
{
    char c;

    while (1)
        if (read(0, &c, 1) < 0 // assume fails only when errno == EINTR
#ifdef SIGWINCH
            || win_size_changed
#endif
            )
        {
#ifdef SIGWINCH
            if (win_size_changed)
            {
                win_size_changed = 0;
                adjust_window();
                redisplay();
            }
#endif
        }
        else
            return c;
}


//
// Executes command using exec.  Returns 1 if the exec went OK,
// otherwise it returns 0.  If closem is true, which is the default,
// we close file descriptors 0, 1 and 2.
//

int execute (const char *file, const char *argv[], int closem)
{
    int status;
    pid_t pid = fork();
    switch(pid)
    {
      case -1:
        return 0;
      case 0:
        if (closem)
        {
            (void)close(0);
            (void)close(1);
            (void)close(2);
        }
        execvp(file, argv);
        //
        // Exec failed.
        //
        exit(1);
      default:
#ifdef NOWAITPID
        while (wait(&status) !=  pid) ;
#else
        waitpid(pid, &status, 0);
#endif
        return status == 0 ? 1 : 0;
    }
}


//
// Execute the passed command using system. If prompt == 1, which is
// the default, we prompt for a key before returning.  We use system so that
// shell and environment variables will be expanded.
//

void exec_with_system (const char *cmd, int prompt)
{
    deinit_screen_and_kbdr();
    unset_signals();

    system(cmd);

    set_signals();
    setraw();

    if (prompt)
        eat_a_character("Press Any Key to Continue");

    enter_cursor_addressing_mode();
    enter_visual_mode();
    enable_keypad();
    synch_display();
}


#ifdef SIGWINCH

#include <sys/ioctl.h>

#ifdef sgi
extern "C" int ioctl (int, int, ...);
#endif

//
// Called to adjust our window after getting a SIGWINCH.
//

static void adjust_window (void)
{
#ifdef TIOCGWINSZ
    struct winsize w;
    if (ioctl(fileno(stdout), TIOCGWINSZ, (char *)&w) == 0 && w.ws_row > 0)
        LI = w.ws_row;
    if (ioctl(fileno(stdout), TIOCGWINSZ, (char *)&w) == 0 && w.ws_col > 0)
        CO = w.ws_col;

    DirList* dl = dir_stack->top();
    //
    // Is current line still on the screen?
    //
    if (dl->savedYPos() >= rows()-2)
    {
        dl->setCurrLine(dir_stack->top()->firstLine());
        dl->saveYXPos(0, goal_column(dir_stack->top()));
    }

    //
    // need to adjust lastLine
    //
    DirLine *ln = dl->firstLine();
    for (int i = 0; i < rows()-2 && ln; i++, ln = ln->next())
        ;
    if (ln)
        dl->setLast(ln->prev());
    else
        dl->setLast(dl->tail());
#endif
}
#endif


#ifdef COMPLETION

//
// This is the sort routine we pass to qsort for sorting the
// table of possible completions.
//

static int string_sort (const void *a, const void *b)
{
    return strcmp(*(char **)a, *(char **)b);
}


//
// Read directory dir for any completions of prefix.  Returns zero on
// error; if prefix is null; if there are no completions; or if the
// completion is the prefix itself.  Else it returns a string in
// volatile space containing the longest possible completion of prefix,
// minus the characters in the prefix. Used only by prompt.
//

char *complete (const char *directory, const char *prefix)
{
    static char *completion = 0;  // our volatile storage
    size_t prefix_length = strlen(prefix);
    const int chunksize = 10;
    int tablesize = chunksize;
    int matches = 0;  // number of matches of prefix in directory
    
    DIR	*dirp = opendir(directory);
    if (dirp == NULL)
        return 0;

    char **completions = new char*[tablesize * sizeof(char *)];

    DIRENT *entry;
    for (entry = readdir(dirp); entry != NULL; entry = readdir(dirp))
    {
        if (strncmp(entry->d_name, prefix, prefix_length) == 0)
        {
            //
            // We have got a match.
            //
            completions[matches] = new char[strlen(entry->d_name) + 1];
            (void)strcpy(completions[matches++], entry->d_name);
            if (matches == tablesize)
            {
                //
                // Grow table.
                //
                tablesize += chunksize;
                char **newtable = new char*[tablesize * sizeof(char *)];
                memcpy(newtable, completions, (tablesize-chunksize)*sizeof(char*));
                delete [] completions;
                completions = newtable;
            }
        }
    }

    //
    // I should really be testing the rc here, but so many of the
    // machines I have tested have this improperly prototyped to void,
    // that it is best not to for the time being.
    //
    (void)closedir(dirp);

    if (matches == 0)
    {
        delete [] completions;
        return 0;
    }

    if (matches == 1)
    {
        if (strcmp(prefix, completions[0]) == 0)
        {
            //
            // If the completion matches our input prefix, we return 0.
            //
            delete [] completions[0];
            delete [] completions;
            return 0;
        }
        delete [] completion;

        completion = new char[strlen(completions[0]) - prefix_length +1];
        (void) strcpy(completion, completions[0] + prefix_length);

        delete [] completions[0];
        delete [] completions;

        return completion;
    }

    qsort(completions, matches, sizeof(char **), string_sort);

    //
    // The completion cannot be longer than the first sorted item.
    //
    size_t maxlen = strlen(completions[0]);
    char *tmp = new char[maxlen + 1];
    (void) strcpy(tmp, prefix);
    size_t i = prefix_length;
    char c = completions[0][i];
    for (; i < maxlen; i++, c = completions[0][i])
    {
        int stop = 0;
        for (int j = 1; j < matches; j++)
        {
            if (c != completions[j][i])
            {
                stop = 1;
                break;
            }
        }
        if (stop)
            break;
        tmp[i] = c;
    }
    tmp[i] = 0;  // nullify the string
    if (strcmp(prefix, tmp) == 0)
    {
        //
        // If the completion matches our input prefix, we return 0.
        //
        delete [] tmp;
        for (i = 0; i < matches; i++)
            delete [] completions[i];
        delete [] completions;

        return 0;
    }
    delete [] completion;
    completion = new char[maxlen - prefix_length + 1];
    (void)strcpy(completion, tmp + prefix_length);

    delete [] tmp;

    for (i = 0; i < matches; i++)
        delete [] completions[i];
    delete [] completions;

    return completion;
}

#endif /*COMPLETION*/


//
// Deletes one char backward while in a prompt. Beeps if there are no
// character to delete.
//
 
static void dispose_of_char (size_t& pos, size_t& space_available, int msglen,
                             char *response, const char *msg)
{
    if (pos == 0)
    {
        ding();
        return;
    }

    backspace();
    DC ? delete_char_at_cursor() : clear_to_end_of_line();
    --pos;
    ++space_available;
    if (space_available == columns())
    {
        //
        // The only way this can happen is if we had previously shifted the
        // response to the left.  Now we must shift the response to the right.
        //
        clear_message_line();
        response[pos] = 0;
        if (pos + msglen < columns())
        {
            //
            // Output message and response-to-date.
            //
            (void) fputs(msg, stdout);
            (void) fputs(response, stdout);
            space_available = columns() - pos - msglen;
        }
        else if (pos < columns())
        {
            //
            // Display the response.
            //
            (void) fputs(response, stdout);
            space_available = columns() - strlen(response);
        }
        else
        {
            //
            // Display the backend of the response
            //
            (void) fputs(&response[pos - columns()/2 + 1], stdout);
            space_available = columns()/2 + 1;
        }
    }
    synch_display();
}


//
// Displays msg prompt and then collects the response.
// The keys of the response are echoed as they are collected.  
// A response of 0 indicates that the command was aborted.
// A response can contain any graphical character.
// C-G will abort out of a prompt;
// Backspace and DELETE work as expected.
// ^U deletes word preceding the cursor.
// ^W delete the complete response.
// Carriage return indicates the end of response.
// Non-graphical characters are ignored.
// If do_completion != 0, with do_completon == 0 by default,
// we do filename completion on the TAB character.
// The response is kept in volatile storage, so if the client
// needs to save it, they must make a copy of it.
//

#ifdef COMPLETION
const char *prompt (const char *msg, int do_completion)
#else
const char *prompt (const char *msg)
#endif
{
    static char *response;  // our volatile storage

    size_t written;         // number of characters written to message line
    const char *abort_msg = "(C-g to Abort) ";
    String nmsg(abort_msg);
    nmsg += msg;
    size_t msglen = nmsg.length();

    move_to_message_line(); // it will have been already cleared by get_key

    if (msglen < columns())
    {
        (void) fputs(nmsg, stdout);
        written = msglen;
    }
    else
    {
        //
        // Leave space for columns/2 + 1 characters.
        //
        (void) fputs((const char *)nmsg + (msglen-columns()/2+1), stdout);
        written = columns()/2 - 1;
    }
    synch_display();

    //
    // We never echo into the last position in the message window.
    //
    size_t space_available = columns() - written; // available spaces in line

    delete [] response;  // release any volatile storage
    response = new char[space_available + 1];
    size_t pos = 0;  // index of next character in response

    char key;
    for (;;)
    {
        if (read(0, &key, 1) < 0 // assume fails only when errno == EINTR
#ifdef SIGWINCH
            || win_size_changed
#endif
            )
        {
#ifdef SIGWINCH
            if (win_size_changed)
            {
                win_size_changed = 0;
                adjust_window();
                redisplay();
            }
#endif
            //
            // On a SIGTSTP the signal handler does the redisplay,
            // so all we have to worry about is getting the prompt
            // redisplayed correctly.
            // Must make sure total output is less than screen width.
            //
            clear_message_line();
            response[pos] = 0;
            if (pos + msglen < columns())
            {
                //
                // Output message and response-to-date.
                //
                (void) fputs(nmsg, stdout);
                (void) fputs(response, stdout);
                space_available = columns() - pos - msglen;
            }
            else if (pos < columns())
            {
                //
                // Display the response.
                //
                (void) fputs(response, stdout);
                space_available = columns() - strlen(response);
            }
            else
            {
                //
                // Display the backend of the response
                //
                (void) fputs(&response[pos - columns()/2 + 1], stdout);
                space_available = columns()/2 + 1;
            }
            synch_display();
        }
        else if (isprint(key))
        {
            //
            // Echo character to message window and wait for another.
            //
            response[pos++] = key;
            space_available--;
            if (!space_available)
            {
                //
                // Need to allocate more room for the response.
                // Note that strlen(response) == pos
                //
                space_available = columns()/2 + 1;
                char *nresponse = new char[pos + space_available + 1];
                response[pos] = 0;  // stringify response
                (void) strcpy(nresponse, response);
                delete [] response;
                response = nresponse;
                //
                // Shift prompt in message window so we
                // always have the end in view to which we are
                // adding characters as they are typed.
                //
                clear_message_line();
                (void) fputs(&response[pos - columns()/2 + 1], stdout);
            }
            else
                putchar(key);
            synch_display();
        }
        else
            switch (key)
            {
#ifdef COMPLETION
              case KEY_TAB:
              {
                  //
                  // Perform filename completion.
                  //
                  if (do_completion == 0 || pos == 0)
                  {
                      ding();
                      break;
                  }
                  
                  response[pos] = 0; // stringify response
                  char *last = strrchr(response, '/');
                  const char *current_directory = dir_stack->top()->name();
                  char *completion;
                  
                  if (response[0] == '~')
                  {
                      //
                      // Expand the tilde before trying the completion,
                      // but there is no use trying the completion if we do not
                      // have a string of the form: "~/.*".
                      //
                      if (response[1] != '/')
                      {
                          ding();
                          break;
                      }
                      String head(response);
                      char *expansion = (char *) expand_tilde(head);
                      char *last = strrchr(expansion, '/');
                      expansion[last - expansion] = 0;
                      completion = complete(expansion, last + 1);
                  }
                  else if (last == NULL)
                      //
                      // Filename is relative to our current directory.
                      //
                      completion = complete(current_directory, response);
                  else
                  {
                      //
                      // The response contains a slash.
                      //
                      if (response[0] == '/')
                      {
                          //
                          // It is an absolute path.
                          //
                          if (pos == 1)
                          {
                              ding();
                              break;
                          }
                          if (response == last)
                              //
                              // The first slash is also the last.
                              //
                              completion = complete("/", last + 1);
                          else
                          {
                              String head(response);
                              head[last - response] = 0;
                              completion = complete(head, last + 1);
                          }
                      }
                      else
                      {
                          //
                          // It is relative to our current directory.
                          //
                          String head(current_directory);
                          head += "/";
                          head += response;
                          head[strlen(current_directory)+last-response+1] = 0;
                          completion = complete(head, last + 1);
                      }
                  }
                  if (completion == 0)
                  {
                      ding();
                      break;
                  }
                  
                  size_t clen = strlen(completion);
                  pos += clen;
                  space_available -= clen;
                  if (space_available > 0)
                  {
                      //
                      // We have got the space to hold the completed filename.
                      //
                      (void) strcat(response, completion);
                      (void) fputs(completion, stdout);
                  }
                  else
                  {
                      //
                      // Allocate more space for response and adjust
                      // message line.
                      //
                      char *nresponse = new char[pos + columns()/2 + 1];
                      (void) strcpy(nresponse, response);
                      (void) strcat(nresponse, completion);
                      delete [] response;
                      response = nresponse;
                      clear_message_line();
                      (void) fputs(&response[pos - columns()/2 + 1], stdout);
                      space_available = columns()/2 + 1;
                  }
                  synch_display();
                  break;
              }

#endif /*COMPLETION*/

              case KEY_CR: // we have the complete response
                response[pos] = 0;
                clear_message_line();
                synch_display();
                return response;

              case KEY_ABORT:  // abort --  reset cursor to previous position
              {
                DirList* dl = dir_stack->top();
                move_cursor(dl->savedYPos(), dl->savedXPos());
                message("Aborted");
                return 0;
              }

              case KEY_DEL:
              case KEY_BKSP: // back up one character
                dispose_of_char(pos, space_available, msglen, response, nmsg);
                break;
              case KEY_CTL_U: // clear line redisplay prompt
                if (pos + msglen >= columns())
                {
                    //
                    // nmsg is not visible
                    //
                    clear_message_line();
                    if (msglen < columns())
                    {
                        (void) fputs(nmsg, stdout);
                        written = msglen;
                    }
                    else
                    {
                        (void) fputs((const char*)nmsg +
                                     (msglen-columns()/2 + 1), stdout);
                        written = columns()/2 - 1;
                    }
                    delete [] response;
                    response = new char[space_available + 1];
                }
                else
                {
                    //
                    // msg is visible
                    //
                    written = msglen < columns() ? msglen : columns()/2 - 1;
                    move_cursor(rows() - 1, written);
                    clear_to_end_of_line();
                }
                synch_display();
                space_available = columns() - written;
                pos = 0;
                break;
              case KEY_CTL_W: // clear word
                {
                    int leading_blanks = response[pos-1] == ' ' ||
                                         response[pos-1] == '\t';
                    do
                    {
                        if (response[pos-1] != ' ' && response[pos-1] != '\t')
                            leading_blanks = 0;

                        if (!leading_blanks && (response[pos-1] == ' ' ||
                                                response[pos-1] == '\t'))
                            break;

                        dispose_of_char(pos, space_available, msglen,
                                        response, nmsg);
                    }
                    while (pos > 0);
                }
                break;
              default: ding(); break; // ignore other characters
            }
    }
}


//
// Returns the number of lines in the DirList currently displayed
// on the screen.
//

int lines_displayed (DirList *dl)
{
    DirLine *ln = dl->firstLine();
    for (int i = 1; ln != dl->lastLine(); i++, ln = ln->next())
        ;
    return i;
}


//
// Prints a message on the last line of the screen. It is up to the
// calling process to put the cursor back where it belongs.  Synchs
// the display.  It can be called as either:
//
//                message(msg);
//           or
//                message(fmt, str);
//
// In the later case it must be the case that the format fmt has exactly
// one % into which the str will be substituted as in the ?printf functions. 
//

int message_window_dirty = 0;

void message (const char *fmt, const char *str)
{
    String msg;          // the complete message to be output

    clear_message_line();

    if (str)
    {
        const char *token = strchr(fmt, '%');
        if (token == 0)
            //
            // This should not happen.  But if it does, let us
            // just print the format fmt.
            //
            msg = fmt;
        else
        {
            msg  = String(fmt, token - fmt);
            msg += str;
            msg += token + 1;
        }
    }
    else
        msg = fmt;

    if (msg.length() < columns())
        (void)fputs(msg, stdout);
    else
        (void)printf("%*.*s", columns() - 1, columns() - 1, (const char *)msg);

    synch_display();
    message_window_dirty = 1;
}


//
// Called both on receipt of certain signals.
//

SIGFUNC(exit_nicely) { deinit_screen_and_kbdr(); exit(1); }

//
// Prints the initial listing screen.  Called  by dired and read_commands
// when rereading the current directory. Adjusts firstLine, lastLine
// and currLine.
//

void initial_listing (DirList *dl)
{
    DirLine *ln = dl->head();
    dl->setFirst(ln);
    dl->setCurrLine(ln);

    clear_display_area();
    
    for (int i = 0; i < rows()-2 && ln; ln = ln->next(), i++)
        display_string(ln->line(), ln->length());

    if (ln)
        dl->setLast(ln->prev());
    else
        dl->setLast(dl->tail());
}


#ifdef SIGWINCH

//
// Set flag indicating window size changed.
//

SIGFUNC(winch)
{
#ifdef sgi
    (void) signal(SIGWINCH, (void (*)(int ...)) SIG_IGN);
#else
    (void) signal(SIGWINCH, SIG_IGN);
#endif
    win_size_changed = 1;
#ifdef sgi
    (void) signal(SIGWINCH, (void (*)(int ...)) winch);
#else
    (void) signal(SIGWINCH, winch);
#endif
}

#endif    


//
// Set up our signal handlers.
//

void set_signals (void)
{
#ifdef sgi
    (void) signal(SIGHUP,  (void (*)(int ...)) exit_nicely);
    (void) signal(SIGINT,  (void (*)(int ...)) exit_nicely);
    (void) signal(SIGQUIT, (void (*)(int ...)) exit_nicely);
    (void) signal(SIGTERM, (void (*)(int ...)) exit_nicely);
#ifdef SIGTSTP
    (void) signal(SIGTSTP, (void (*)(int ...)) termstop);
#endif
#ifdef SIGWINCH
    (void) signal(SIGWINCH, (void (*)(int ...)) winch);
#endif
#else
    (void) signal(SIGHUP,  exit_nicely);
    (void) signal(SIGINT,  exit_nicely);
    (void) signal(SIGQUIT, exit_nicely);
    (void) signal(SIGTERM, exit_nicely);
#ifdef SIGTSTP
    (void) signal(SIGTSTP, termstop);
#endif
#ifdef SIGWINCH
    (void) signal(SIGWINCH, winch);
#endif
#endif
}


//
// Set signals back to defaults.
//

void unset_signals (void)
{
    (void) signal(SIGHUP,  SIG_DFL);
    (void) signal(SIGINT,  SIG_DFL);
    (void) signal(SIGQUIT, SIG_DFL);
    (void) signal(SIGTERM, SIG_DFL);
#ifdef SIGTSTP
    (void) signal(SIGTSTP, SIG_DFL);
#endif
#ifdef SIGWINCH
    (void) signal(SIGWINCH, SIG_DFL);
#endif
}


//
// Shifts the current line in DirList left until its tail is visible.
//
void leftshift_current_line (DirList *dl)
{
    int inc = dl->currLine()->length()-columns()+1;
    move_cursor(dl->savedYPos(), 0);
    clear_to_end_of_line();
    display_string(&(dl->currLine()->line())[inc],columns()-1);
    dl->saveYXPos(dl->savedYPos(), max(goal_column(dl)-inc, 0));
    move_cursor(dl->savedYPos(), dl->savedXPos());
}


//
// Rightshifts current line to "natural" position.
//
void rightshift_current_line (DirList *dl)
{
    move_cursor(dl->savedYPos(), 0);
    clear_to_end_of_line();
    display_string(dl->currLine()->line(), dl->currLine()->length());
    dl->saveYXPos(dl->savedYPos(), goal_column(dl));
    move_cursor(dl->savedYPos(), dl->savedXPos());
}
